import json
from qtpy.QtWidgets import QWidget, QPlainTextEdit, QVBoxLayout, QHBoxLayout, QGridLayout, QPushButton, QApplication, QLabel
from qtpy.QtCore import Qt
from qtpy.QtGui import QFont

class MacroWindow(QWidget):
    """
    A replica of EDM's "Show Macros" display
    """
    def __init__(self, parent=None):
        super().__init__(parent, Qt.Window)
        self.setWindowTitle("Macros")
        self.setFixedWidth(750)
        self.setFixedHeight(500)

        self.text_box = QPlainTextEdit()
        self.text_box.setReadOnly(True)
        self.font = QFont()
        self.font.setPointSize(16)
        self.text_box.setFont(self.font)
        self.text_box.setPlainText("Macros:")

        self.macros = {}
        self.init_macros()
        self.populate_macros()

        self.label = QLabel()
        self.label.setText('')
        self.label.setMinimumSize(100, 25)

        self.copy_button = QPushButton()
        self.copy_button.setCheckable(False)
        self.copy_button.setMinimumSize(75, 25)
        self.copy_button.setText("Copy Macros as JSON")
        self.copy_button.clicked.connect(self.copy_macros)

        self.clipboard = QApplication.clipboard()

        self.vBoxLayout = QVBoxLayout()
        self.vBoxLayout.addWidget(self.text_box)
        self.hBoxLayout = QHBoxLayout()
        self.hBoxLayout.addWidget(self.label, alignment=Qt.AlignRight)
        self.hBoxLayout.addWidget(self.copy_button, alignment=Qt.AlignRight)
        self.hBoxLayout.insertStretch(0)
        self.vBoxLayout.addLayout(self.hBoxLayout)

        self.setLayout(self.vBoxLayout)

    def init_macros(self):
        if self.parent() is not None:
            display_widget = self.parent().display_widget()
            if display_widget is not None:
                self.macros = display_widget.macros()

    def populate_macros(self):
        for macro, value in self.macros.items():
            line = "  {}={}".format(macro, value)
            self.text_box.appendPlainText(line)

    def copy_macros(self):
        self.clipboard.setText(json.dumps(self.macros))
        self.label.setText('Macros copied to clipboard!')



"""Module to define a parent qtdesigner plugin class.

Please continue to name your qtdesigner plugin modules using the convention
modulename.py          <--- defines the widget
modulename_qtplugin.py <--- imports this module + the widget

However, a majority of the builtin plugins are defined in qtplugins.py
adjacent to this module.

NOTE: PyDMDesignerPlugin is a valid plugin, so designer will try to pick it up
      and instantiate it if you import it into another module's namespace. You
      will need to avoid having it present in the global namespace of any
      module that defines a qtplugin.

If you do not heed this warning, you will get a one-line traceback:
TypeError: __init__() takes exactly 3 arguments (1 given)
for each PyDMDesignerPlugin that Qt Designer tries to use. This will not
affect any of your widgets, but it will be annoying.

"""
import enum
import inspect
import logging
from typing import Dict, List, Optional, Type

import entrypoints
from qtpy import QtCore, QtDesigner, QtGui, QtWidgets

from .. import config
from ..qtdesigner import DesignerHooks
from .qtplugin_extensions import PyDMExtensionFactory

logger = logging.getLogger(__name__)


class WidgetCategory(str, enum.Enum):
    """Categories for PyDM Widgets in the Qt Designer."""
    #: Widgets which can contain other widgets.
    CONTAINER = "PyDM Container Widgets"
    #: Widgets which contain displays.
    DISPLAY = "PyDM Display Widgets"
    #: Widgets which take user input.
    INPUT = "PyDM Input Widgets"
    #: Widgets which plot data.
    PLOT = "PyDM Plot Widgets"
    #: Widgets which draw things.
    DRAWING = "PyDM Drawing Widgets"
    #: Widgets which don't fit into any other category.
    MISC = "PyDM Misc. Widgets"


def qtplugin_factory(
    cls: Type[QtWidgets.QWidget],
    is_container: bool = False,
    group: str = "PyDM Widgets",
    extensions: Optional[List[Type]] = None,
    icon: Optional[QtGui.QIcon] = None
) -> Type["PyDMDesignerPlugin"]:
    """
    Helper function to create a generic PyDMDesignerPlugin class.

    Parameters
    ----------
    cls : QWidget subclass
        The widget class.

    is_container : bool, optional
        True if this widget can contain other widgets (as in a Frame).
        This will also affect whether or not the widget can be used as the
        top-level widget of a display. Defaults to False.

    extensions : list of extension classes, optional
        Extension classes to use with the widget.

    icon : QtGui.QIcon, optional
        An icon to use with the widget in the designer.  Consider using
        the PyDM-provided fontawesome ``ifont`` here for simplicity.
    """

    class Plugin(PyDMDesignerPlugin):
        __doc__ = "PyDMDesigner Plugin for {}".format(cls.__name__)

        def __init__(self):
            super(Plugin, self).__init__(cls, is_container, group, extensions,
                                         icon)

    return Plugin


class PyDMDesignerPlugin(QtDesigner.QPyDesignerCustomWidgetPlugin):
    """
    Parent class to standardize how pydm plugins are accessed in qt designer.
    All functions have default returns that can be overriden as necessary.
    """

    def __init__(
        self,
        cls: Type[QtWidgets.QWidget],
        is_container: bool = False,
        group: str = "PyDM Widgets",
        extensions: Optional[List[Type]] = None,
        icon: Optional[QtGui.QIcon] = None,
    ):
        """
        Set up the plugin using the class info in cls

        :param cls: Class of the widget to use
        :type cls:  QWidget
        """
        QtDesigner.QPyDesignerCustomWidgetPlugin.__init__(self)
        self.initialized = False
        self.is_container = is_container
        self.cls = cls
        self._group = group

        if icon is None and QtWidgets.QApplication.instance() is not None:
            get_designer_icon = getattr(cls, "get_designer_icon", None)
            if get_designer_icon is not None:
                icon = get_designer_icon()

            if icon is None:
                icon = QtGui.QIcon()

        self._icon = icon
        self.extensions = extensions
        self.manager = None

    def initialize(self, core):
        """
        Override this function if you need special initialization instructions.
        Make sure you don't neglect to set the self.initialized flag to True
        after a successful initialization.

        :param core: form editor interface to use in the initialization
        :type core:  QDesignerFormEditorInterface
        """
        if self.initialized:
            return

        self.initialized = True

        designer_hooks = DesignerHooks()
        designer_hooks.form_editor = core

        if self.extensions is not None and len(self.extensions) > 0:
            self.manager = core.extensionManager()
            if self.manager:
                factory = PyDMExtensionFactory(parent=self.manager)
                self.manager.registerExtensions(
                    factory,
                    'org.qt-project.Qt.Designer.TaskMenu')  # Qt5

    def isInitialized(self):
        """
        Return True if initialize function has been called successfully.
        """
        return self.initialized

    def createWidget(self, parent):
        """
        Instantiate a widget with the given parent.

        :param parent: Parent widget of instantiated widget
        :type parent:  QWidget
        """
        w = self.cls(parent=parent)
        try:
            setattr(w, "extensions", self.extensions)
            w.init_for_designer()
        except (AttributeError, NameError):
            pass
        return w

    def name(self):
        """
        Return the class name of the widget.
        """
        return self.cls.__name__

    def group(self):
        """
        Return a common group name so all PyDM Widgets are together in
        Qt Designer.
        """
        return self._group

    def toolTip(self):
        """
        A short description to pop up on mouseover. If we leave this as an
        empty string, we'll have no tooltip by default and can override this
        on a case-by-case basis.
        """
        return ""

    def whatsThis(self):
        """
        A longer description of the widget for Qt Designer. By default, this
        is the entire class docstring.
        """
        return ""

    def isContainer(self):
        """
        Return True if this widget can contain other widgets.
        """
        return self.is_container

    def icon(self):
        """
        Return a QIcon to represent this widget in Qt Designer.
        """
        return QtGui.QIcon(self._icon.pixmap(QtCore.QSize(32, 32)))

    def domXml(self):
        """
        XML Description of the widget's properties.
        """
        return (
            "<widget class=\"{0}\" name=\"{0}\">\n"
            " <property name=\"toolTip\" >\n"
            "  <string>{1}</string>\n"
            " </property>\n"
            "</widget>\n"
        ).format(self.name(), self.toolTip())

    def includeFile(self):
        """
        Include the class module for the generated qt code
        """
        return self.cls.__module__


def create_designer_widget_from_widget(
    widget_cls: Type[QtWidgets.QWidget],
) -> Type[PyDMDesignerPlugin]:
    """
    Get a designable widget class.

    Accepts either user-provided :class:`PyDMDesignerPlugin` subclasses or
    :class:`QWidget` subclasses.

    In the case of :class:`QWidget` subclasses, designer-specific settings
    may be specified on a class attribute.  These arguments should match
    what :func:`qtplugin_factory` expects.
    """
    if not inspect.isclass(widget_cls):
        raise ValueError(
            f"Expected a class, got a {type(widget_cls).__name__}"
        )
    if issubclass(widget_cls, PyDMDesignerPlugin):
        return widget_cls
    if issubclass(widget_cls, QtWidgets.QWidget):
        designer_kwargs = getattr(
            widget_cls, "_qt_designer_", None
        ) or {}
        return qtplugin_factory(widget_cls, **designer_kwargs)

    raise ValueError(
        f"Expected a PyDMDesignerPlugin or a QWidget subclass, "
        f"got a {type(widget_cls).__name__}"
    )


def get_widgets_from_entrypoints(
    key: str = config.ENTRYPOINT_WIDGET
) -> Dict[str, Type[PyDMDesignerPlugin]]:
    """
    Get all widgets from entrypoint definitions.

    Parameters
    ----------
    key : str, optional
        The entrypoint key.  Defaults to ``pydm.config.ENTRYPOINT_WIDGET``.

    Returns
    -------
    widgets : dict
        Dictionary of class name to ``PyDMDesignerPlugin`` subclass.
    """
    widgets = {}
    for entry in entrypoints.get_group_all(key):
        logger.debug("Found widget from entrypoint: %s", entry.name)
        try:
            plugin_cls = entry.load()
        except Exception as ex:
            logger.exception(
                "Failed to load %s entry %s: %s",
                key, entry.name, ex
            )
            continue

        try:
            designer_widget = create_designer_widget_from_widget(
                plugin_cls
            )
        except Exception:
            logger.warning(
                "Invalid widget class specified in entrypoint "
                "%s: %s",
                entry.name, plugin_cls
            )
        else:
            widgets[entry.name] = designer_widget

    return widgets



import json
import itertools
from collections import OrderedDict
import numpy as np
from qtpy.QtGui import QColor
from qtpy.QtCore import Slot, Property, Qt
from .baseplot import BasePlot, NoDataError, BasePlotCurveItem
from .channel import PyDMChannel
from ..utilities import remove_protocol


DEFAULT_BUFFER_SIZE = 1200
MINIMUM_BUFFER_SIZE = 2


class EventPlotCurveItem(BasePlotCurveItem):
    _channels = ('channel')

    def __init__(self, addr, y_idx, x_idx, bufferSizeChannelAddress=None, **kws):
        self.channel = None
        self.address = addr
        self.x_idx = x_idx
        self.y_idx = y_idx
        self.connected = False
        if kws.get('name') is None:
            kws['name'] = ""
        self.bufferSizeChannel = None
        self.bufferSizeChannel_connected = False
        self._bufferSize = DEFAULT_BUFFER_SIZE
        self.data_buffer = np.zeros((2, self._bufferSize),
                                    order='f', dtype=float)
        self.points_accumulated = 0
        if 'symbol' not in kws.keys():
            kws['symbol'] = 'o'
        if 'lineStyle' not in kws.keys():
            kws['lineStyle'] = Qt.NoPen
        super(EventPlotCurveItem, self).__init__(**kws)
        self.bufferSizeChannelAddress = bufferSizeChannelAddress

    def to_dict(self):
        """
        Serialize this curve into a dictionary.

        Returns
        -------
        OrderedDict
            Representation with values for all properties
            needed to recreate this curve.
        """
        dic_ = OrderedDict([("channel", self.address),
                            ("y_idx", self.y_idx),
                            ("x_idx", self.x_idx)])
        dic_.update(super(EventPlotCurveItem, self).to_dict())
        dic_['buffer_size'] = self.getBufferSize()
        dic_["bufferSizeChannelAddress"] = self.bufferSizeChannelAddress
        return dic_

    @property
    def address(self):
        """
        The address of the channel used to get the data.

        Returns
        -------
        str
            The address of the channel used to get the x axis data.
        """
        if self.channel is None:
            return None
        return self.channel.address

    @address.setter
    def address(self, new_address):
        """
        The address of the channel used to get the x axis data.

        Parameters
        -------
        new_address: str
        """
        if new_address is None or len(str(new_address)) < 1:
            self.channel = None
            return
        self.channel = PyDMChannel(
            address=new_address,
            connection_slot=self.connectionStateChanged,
            value_slot=self.receiveValue)

    @Slot(bool)
    def connectionStateChanged(self, connected):
        self.connected = connected

    @Slot(np.ndarray)
    def receiveValue(self, new_data):
        """
        Handler for new data.  This method is usually called by a PyDMChannel
        when it updates.  You can call this yourself to inject data into the curve.

        Parameters
        ----------
        new_data: numpy.ndarray
            A new array of values.
        """
        if new_data is None:
            return
        if self.x_idx is None or self.y_idx is None:
            return
        if not isinstance(self.x_idx, int) or not isinstance(self.y_idx, int):
            """ The x_idx and y_idx typing is made this late so that macros can
            can be used alongside regular indexing. """
            self.x_idx = int(self.x_idx)
            self.y_idx = int(self.y_idx)
        if len(new_data) <= self.x_idx or len(new_data) <= self.y_idx:
            return
        self.data_buffer = np.roll(self.data_buffer, -1)
        self.data_buffer[0, -1] = new_data[self.x_idx]
        self.data_buffer[1, -1] = new_data[self.y_idx]
        if self.points_accumulated < self._bufferSize:
            self.points_accumulated = self.points_accumulated + 1
        self.data_changed.emit()

    def initialize_buffer(self):
        self.points_accumulated = 0
        self.data_buffer = np.zeros((2, self._bufferSize),
                                    order='f', dtype=float)

    def getBufferSize(self):
        return int(self._bufferSize)

    def setBufferSize(self, value):
        if self._bufferSize != int(value):
            self._bufferSize = max(int(value), MINIMUM_BUFFER_SIZE)
            self.initialize_buffer()

    def resetBufferSize(self):
        if self._bufferSize != DEFAULT_BUFFER_SIZE:
            self._bufferSize = DEFAULT_BUFFER_SIZE
            self.initialize_buffer()

    @property
    def bufferSizeChannelAddress(self):
        """
        The address of the channel used to get the buffer size.

        Returns
        -------
        str
            The address of the channel used to get the buffer size.
        """
        if self.bufferSizeChannel is None:
            return None
        return self.bufferSizeChannel.address

    @bufferSizeChannelAddress.setter
    def bufferSizeChannelAddress(self, new_address):
        """
        The address of the channel used to get the buffer size.

        Parameters
        ----------
        new_address: str
        """
        if new_address is None or len(str(new_address)) < 1:
            self.bufferSizeChannel = None
            return
        if self.bufferSizeChannel is not None and self.bufferSizeChannel.address == new_address:
            return
        self.bufferSizeChannel = PyDMChannel(
            address=new_address,
            connection_slot=self.bufferSizeConnectionStateChanged,
            value_slot=self.bufferSizeChannelValueReceiver)
        self.bufferSizeChannel.connect()

    @Slot(bool)
    def bufferSizeConnectionStateChanged(self, connected):
        self.bufferSizeChannel_connected = connected

    @Slot(int)
    def bufferSizeChannelValueReceiver(self, value):
        """
        Handler for change in buffer size.  This method is usually called by a PyDMChannel
        when it updates.  You can call this yourself to set or change the buffer size.

        Parameters
        ----------
        value: int
            A new value for the buffer size.
        """
        if value is None:
            return
        self.setBufferSize(value)
        self.update_buffer()

    def redrawCurve(self):
        """
        Called by the curve's parent plot whenever the curve needs to be
        re-drawn with new data.
        """
        self.setData(x=self.data_buffer[0, -self.points_accumulated:].astype(float),
                     y=self.data_buffer[1, -self.points_accumulated:].astype(float))

    def limits(self):
        """
        Get the limits of the data for this curve.

        Returns
        -------
        tuple
            A nested tuple of limits: ((xmin, xmax), (ymin, ymax))
        """
        if self.points_accumulated == 0:
            raise NoDataError("Curve has no data, cannot determine limits.")
        x_data = self.data_buffer[0, -self.points_accumulated:]
        y_data = self.data_buffer[1, -self.points_accumulated:]
        return ((float(np.amin(x_data)), float(np.amax(x_data))),
                (float(np.amin(y_data)), float(np.amax(y_data))))

    def channels(self):
        return [self.channel]


class PyDMEventPlot(BasePlot):
    """
    PyDMEventPlot is a widget to plot one scalar value against another.
    All of the values arrive in a single event-built array, and indices are
    used to identify which values to plot.  Multiple scalar pairs can be 
    plotted on the same plot.  Each pair has a buffer which stores previous
    values.  All values in the buffer are drawn.  The buffer size for each
    pair is user configurable.

    Parameters
    ----------
    parent : optional
        The parent of this widget.
    channel : optional
        A string with the address for a channel.  This channel should produce
        an array of data for each event.  The index parameters are used to
        select data from this array to be plotted.
    init_x_indices: optional
        init_x_indices can be None, an integer, or a list of integers possibly
        containing None.  These indices are used to select values from the
        data array provided by channel to be plotted.  In the event that
        an index is None, the curve will be created, but no plotting will
        occur until an integer index is set.  If lists are specified for
        both init_x_indices and init_y_indices, they both must have the same
        length. If a single x index was specified, and a list of y indices is
        specified, all of the selected y data will be plotted against the same x.
    init_y_indices: optional
        init_y_indices can be None, an integer, or a list of integers possibly
        containing None.  These indices are used to select values from the
        data array provided by channel to be plotted.  In the event that
        an index is None, the curve will be created, but no plotting will
        occur until an integer index is set.  If lists are specified for
        both init_x_indices and init_y_indices, they both must have the same
        length. If a single x index was specified, and a list of y indices is
        specified, all of the selected y data will be plotted against the same x.
    background: optional
        The background color for the plot. Accepts any arguments that
        pyqtgraph.mkColor will accept.
    """
    def __init__(self, parent=None, channel=None, init_x_indices=[], init_y_indices=[],
                 background='default'):
        super(PyDMEventPlot, self).__init__(parent, background)
        # If the user supplies a single integer instead of a list,
        # wrap it in a list.
        if isinstance(init_x_indices, int):
            init_x_indices = [init_x_indices]
        if isinstance(init_y_indices, int):
            init_y_indices = [init_y_indices]
        if init_y_indices is None:
            init_y_indices = []
        if init_x_indices is None or len(init_x_indices) == 0:
            init_x_indices = list(itertools.repeat(None,
                                                   len(init_y_indices)))
        if len(init_x_indices) == 1:
            init_x_indices = init_x_indices*len(init_y_indices)
        if len(init_x_indices) != len(init_y_indices):
            raise ValueError("If lists are provided for both X and Y " +
                             "indices, they must be the same length.")
        # self.index_pairs is an ordered dictionary that is keyed on a
        # (x_idx, y_idx) tuple, with EventPlotCurveItem values.
        # It gets populated in self.addChannel().
        self.index_pairs = OrderedDict()
        init_index_pairs = zip(init_x_indices, init_y_indices)
        for (x_idx, y_idx) in init_index_pairs:
            self.addChannel(channel=channel, y_idx=y_idx, x_idx=x_idx)
        self._needs_redraw = True

    def initialize_for_designer(self):
        # If we are in Qt Designer, don't update the plot continuously.
        # This function gets called by PyDMTimePlot's designer plugin.
        pass

    def addChannel(self, channel=None, y_idx=None, x_idx=None, name=None,
                   color=None, lineStyle=None, lineWidth=None,
                   symbol='o', symbolSize=5, buffer_size=None,
                   yAxisName=None, bufferSizeChannelAddress=None):
        """
        Add a new curve to the plot.  In addition to the arguments below,
        all other keyword arguments are passed to the underlying
        pyqtgraph.PlotDataItem used to draw the curve.

        Parameters
        ----------
        channel: str
            The channel for the curve data.
        y_idx: int
            The index for the y data for the curve.
        x_idx: int, optional
            The index for the x data for the curve.
        name: str, optional
            A name for this curve.  The name will be used in the plot legend.
        color: str or QColor, optional
            A color for the line of the curve.  If not specified, the plot will
            automatically assign a unique color from a set of default colors.
        lineStyle: int, optional
            Style of the line connecting the data points.
            0 means no line (scatter plot).
        lineWidth: int, optional
            Width of the line connecting the data points.
        buffer_size: int, optional
            number of points to keep in the buffer.
        bufferSizeChannelAddress : str, optional
            The name of a channel that defines the buffer size (int).
        symbol: str or None, optional
            Which symbol to use to represent the data.
        symbol: int, optional
            Size of the symbol.
        yAxisName : str, optional
            The name of the y axis to associate with this curve. Will be created if it
            doesn't yet exist
        """
        plot_opts = {}
        plot_opts['symbol'] = symbol
        if symbolSize is not None:
            plot_opts['symbolSize'] = symbolSize
        if lineStyle is not None:
            plot_opts['lineStyle'] = lineStyle
        if lineWidth is not None:
            plot_opts['lineWidth'] = lineWidth
        curve = EventPlotCurveItem(addr=channel,
                                   y_idx=y_idx,
                                   x_idx=x_idx,
                                   name=name,
                                   color=color,
                                   yAxisName=yAxisName,
                                   bufferSizeChannelAddress=bufferSizeChannelAddress,
                                   **plot_opts)
        if buffer_size is not None:
            curve.setBufferSize(buffer_size)
        self.index_pairs[(x_idx, y_idx)] = curve
        self.addCurve(curve, curve_color=color, y_axis_name=yAxisName)
        curve.data_changed.connect(self.set_needs_redraw)

    def removeChannel(self, curve):
        """
        Remove a curve from the plot.

        Parameters
        ----------
        curve: EventPlotCurveItem
            The curve to remove.
        """
        self.removeCurve(curve)

    def removeChannelAtIndex(self, index):
        """
        Remove a curve from the plot, given an index
        for a curve.

        Parameters
        ----------
        index: int
            Index for the curve to remove.
        """
        curve = self._curves[index]
        self.removeChannel(curve)

    @Slot()
    def set_needs_redraw(self):
        self._needs_redraw = True

    @Slot()
    def redrawPlot(self):
        """
        Request a redraw from each curve in the plot.
        Called by curves when they get new data.
        """
        if not self._needs_redraw:
            return
        for curve in self._curves:
            curve.redrawCurve()
        self._needs_redraw = False

    def clearCurves(self):
        """
        Remove all curves from the plot.
        """
        super(PyDMEventPlot, self).clear()

    def getCurves(self):
        """
        Get a list of json representations for each curve.
        """
        return [json.dumps(curve.to_dict()) for curve in self._curves]

    def setCurves(self, new_list):
        """
        Replace all existing curves with new ones.  This function
        is mostly used as a way to load curves from a .ui file, and
        almost all users will want to add curves through addChannel,
        not this method.

        Parameters
        ----------
        new_list: list
            A list of json strings representing each curve in the plot.
        """
        try:
            new_list = [json.loads(str(i)) for i in new_list]
        except ValueError as e:
            print("Error parsing curve json data: {}".format(e))
            return
        self.clearCurves()
        for d in new_list:
            color = d.get('color')
            if color:
                color = QColor(color)
            self.addChannel(channel=d['channel'],
                            y_idx=d['y_idx'], x_idx=d['x_idx'],
                            name=d.get('name'), color=color,
                            lineStyle=d.get('lineStyle'),
                            lineWidth=d.get('lineWidth'),
                            symbol=d.get('symbol'),
                            symbolSize=d.get('symbolSize'),
                            buffer_size=d.get('buffer_size'),
                            bufferSizeChannelAddress=d.get('bufferSizeChannelAddress'),
                            yAxisName=d.get('yAxisName'))

    curves = Property("QStringList", getCurves, setCurves, designable=False)

    def channels(self):
        """
        Returns the list of channels used by all curves in the plot.

        Returns
        -------
        list
        """
        chans = []
        chans.extend([curve.channel for curve in self._curves])
        chans.extend([curve.bufferSizeChannel
                      for curve in self._curves
                      if curve.bufferSizeChannel is not None])
        return chans

    # The methods for autoRangeX, minXRange, maxXRange, autoRangeY, minYRange,
    # and maxYRange are all defined in BasePlot, but we don't expose them as
    # properties there, because not all plot subclasses necessarily want
    # them to be user-configurable in Designer.
    autoRangeX = Property(bool, BasePlot.getAutoRangeX,
                          BasePlot.setAutoRangeX, BasePlot.resetAutoRangeX,
                          doc="""
Whether or not the X-axis automatically rescales to fit the data.
If true, the values in minXRange and maxXRange are ignored.""")

    minXRange = Property(float, BasePlot.getMinXRange,
                         BasePlot.setMinXRange, doc="""
Minimum X-axis value visible on the plot.""")

    maxXRange = Property(float, BasePlot.getMaxXRange,
                         BasePlot.setMaxXRange, doc="""
Maximum X-axis value visible on the plot.""")

    autoRangeY = Property(bool, BasePlot.getAutoRangeY,
                          BasePlot.setAutoRangeY, BasePlot.resetAutoRangeY,
                          doc="""
Whether or not the Y-axis automatically rescales to fit the data.
If true, the values in minYRange and maxYRange are ignored.""")

    minYRange = Property(float, BasePlot.getMinYRange,
                         BasePlot.setMinYRange, doc="""
Minimum Y-axis value visible on the plot.""")

    maxYRange = Property(float, BasePlot.getMaxYRange,
                         BasePlot.setMaxYRange, doc="""
Maximum Y-axis value visible on the plot.""")



# Unit Tests for the PyDMCheckbox Widget

import pytest
from ...widgets.checkbox import PyDMCheckbox


# --------------------
# POSITIVE TEST CASES
# --------------------

@pytest.mark.parametrize("init_channel", [
    "CA://MTEST",
    "",
    None,
])
def test_construct(qtbot, init_channel):
    """
    Test the widget construct.

    Expectations:
    The widget's initial state is unchecked.

    Parameters
    ----------
     qtbot : fixture
        Window for widget testing
    init_channel : str
        The data channel to be used by the widget
    """
    pydm_checkbox = PyDMCheckbox(init_channel=init_channel)
    qtbot.addWidget(pydm_checkbox)

    assert not pydm_checkbox.isChecked()


@pytest.mark.parametrize("init_checked_status, new_value", [
    (True, 0),
    (True, -1),
    (False, 0),
    (False, -66),
    (True, 1),
    (False, 1),
    (True, 999),
    (False, 999),
])
def test_value_changed(qtbot, signals, init_checked_status, new_value):
    """

    Test the widget's checked status when the value changes via the data channel.

    Expectations:

    1. If the channel data value is larger than 0, the widget is checked
    2. The widget is unchecked if the channel value is 0

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    init_checked_status : bool
        True if the widget is initially checked; False otherwise
    new_value : int
        The value sent to the widget from the channel
    """
    pydm_checkbox = PyDMCheckbox()
    qtbot.addWidget(pydm_checkbox)

    pydm_checkbox.setChecked(init_checked_status)

    signals.send_value_signal[type(new_value)].connect(pydm_checkbox.channelValueChanged)
    signals.send_value_signal[type(new_value)].emit(new_value)

    assert pydm_checkbox.isChecked() if new_value > 0 else not pydm_checkbox.isChecked()


@pytest.mark.parametrize("is_checked", [
    True,
    False,
    None,
])
def test_send_value(qtbot, signals, is_checked):
    """
    Test the data sent from the widget to the channel when the widget is checked or unchecked.

    Expectations:

    1. If the widget is checked, it will send out 1 as the value to the channel
    2. If the widget is unchecked, it will send out 0 as the value to the channel

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    signals : fixture
        The signals fixture, which provides access signals to be bound to the appropriate slots
    is_checked : bool
        True if the widget is checked before sending out the value; False if it is not checked
    """
    pydm_checkbox = PyDMCheckbox()
    qtbot.addWidget(pydm_checkbox)

    pydm_checkbox.send_value_signal[int].connect(signals.receiveValue)
    pydm_checkbox.send_value(is_checked)

    assert signals.value == 1 if is_checked else signals.value == 0




import json
import logging
import re
from typing import Any, List, Optional

from qtpy import QtCore, QtDesigner, QtWidgets

from ..utilities import copy_to_clipboard, get_clipboard_text
from ..utilities.macro import parse_macro_string

logger = logging.getLogger(__name__)


def update_property_for_widget(widget: QtWidgets.QWidget, name: str, value):
    """Update a Property for the given widget in the designer."""
    formWindow = QtDesigner.QDesignerFormWindowInterface.findFormWindow(widget)
    logger.info("Updating %s.%s = %s", widget.objectName(), name, value)
    if formWindow:
        formWindow.cursor().setProperty(name, value)
    else:
        setattr(widget, name, value)


class DictionaryTable(QtWidgets.QTableWidget):
    def __init__(self, dictionary=None, *args, parent=None, **kwargs):
        super().__init__(*args, parent=parent, **kwargs)

        self.setColumnCount(2)
        self.setMinimumSize(300, 150)
        self.setHorizontalHeaderLabels(["Key", "Value"])

        self.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self._context_menu)
        self.dictionary = dictionary

    def _context_menu(self, pos):
        self.menu = QtWidgets.QMenu(self)
        item = self.itemAt(pos)
        if item is not None:
            def copy(*_):
                copy_to_clipboard(item.text())

            copy_action = self.menu.addAction(f"&Copy: {item.text()}")
            copy_action.triggered.connect(copy)

            clipboard_text = get_clipboard_text() or ""

            def paste(*_):
                item.setText(clipboard_text)

            paste_action = self.menu.addAction(
                f"&Paste: {clipboard_text[:100]}"
            )
            paste_action.triggered.connect(paste)

            def delete_row(*_):
                self.removeRow(item.row())

            delete_row_action = self.menu.addAction("&Delete row...")
            delete_row_action.triggered.connect(delete_row)

        self.menu.addSeparator()

        def add_row(*_):
            row = self.rowCount()
            self.setRowCount(row + 1)
            self.setItem(row, 0, QtWidgets.QTableWidgetItem(""))
            self.setItem(row, 1, QtWidgets.QTableWidgetItem(""))

        add_row_action = self.menu.addAction("&Add row...")
        add_row_action.triggered.connect(add_row)
        self.menu.exec_(self.mapToGlobal(pos))

    @property
    def dictionary(self) -> dict:
        items = [
            (self.item(row, 0), self.item(row, 1))
            for row in range(self.rowCount())
        ]
        key_value_pairs = [
            (key.text() if key else "", value.text() if value else "")
            for key, value in items
        ]
        return {
            key.strip(): value
            for key, value in key_value_pairs
        }

    @dictionary.setter
    def dictionary(self, dct):
        dct = dct or {}
        self.setRowCount(len(dct))
        for row, (key, value) in enumerate(dct.items()):
            self.setItem(row, 0, QtWidgets.QTableWidgetItem(key))
            self.setItem(row, 1, QtWidgets.QTableWidgetItem(value))

        self.resizeColumnsToContents()
        self.resizeRowsToContents()


class StringListTable(QtWidgets.QTableWidget):
    def __init__(self, values=None, *args, parent=None, **kwargs):
        super().__init__(*args, parent=parent, **kwargs)

        self.setColumnCount(1)
        self.setMinimumSize(300, 150)
        self.setHorizontalHeaderLabels(["Value"])

        self.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self._context_menu)
        self.values

        # we let user swap points by dragging one on-top of the other
        self.setDragEnabled(True)
        self.setAcceptDrops(True)
        self.drag_source_row = None
    
    def startDrag(self, event):
        self.drag_source_row = self.currentRow()
        # call super() since we use the default dragging functionality
        super().startDrag(event)

    def dropEvent(self, event):
        # don't call super() here, functionality messes with our swapping!
        source_row = self.drag_source_row
        target_row = self.indexAt(event.pos()).row()
        if target_row >= 0 and source_row != target_row:
            self.swap_rows(source_row, target_row)
        
    def swap_rows(self, row1, row2):
        item1 = self.takeItem(row1, 0)
        item2 = self.takeItem(row2, 0)
        self.setItem(row1, 0, item2)
        self.setItem(row2, 0, item1)

    def _context_menu(self, pos):
        self.menu = QtWidgets.QMenu(self)
        item = self.itemAt(pos)
        if item is not None:
            def copy(*_):
                copy_to_clipboard(item.text())

            copy_action = self.menu.addAction(f"&Copy: {item.text()}")
            copy_action.triggered.connect(copy)

            clipboard_text = get_clipboard_text()

            def paste(*_):
                item.setText(clipboard_text)

            paste_action = self.menu.addAction(f"&Paste: {clipboard_text}")
            paste_action.triggered.connect(paste)

            def delete_row(*_):
                self.removeRow(item.row())

            delete_row_action = self.menu.addAction("&Delete row...")
            delete_row_action.triggered.connect(delete_row)

        self.menu.addSeparator()

        def add_row(*_):
            row = self.rowCount()
            self.setRowCount(row + 1)
            self.setItem(row, 0, QtWidgets.QTableWidgetItem(""))

        add_row_action = self.menu.addAction("&Add row...")
        add_row_action.triggered.connect(add_row)
        self.menu.exec_(self.mapToGlobal(pos))

    @property
    def values(self) -> list:
        items = [self.item(row, 0) for row in range(self.rowCount())]
        return [
            item.text().strip()
            for item in items
            if item is not None
        ]

    @values.setter
    def values(self, values):
        values = values or []
        self.setRowCount(len(values))
        for row, value in enumerate(values):
            self.setItem(row, 0, QtWidgets.QTableWidgetItem(str(value)))

        self.resizeColumnsToContents()
        self.resizeRowsToContents()


class _PropertyHelper:
    def __init__(self, *args, property_widget, property_name, **kwargs):
        super().__init__(*args, **kwargs)
        self._property_name = property_name
        self._property_widget = property_widget

        value = None
        try:
            value = self.value_from_widget
            self.set_value_from_widget(
                widget=self._property_widget,
                attr=self._property_name,
                value=value,
            )
        except Exception:
            logger.exception(
                "Failed to set helper widget %s state from %s=%s",
                type(self).__name__,
                self._property_name,
                value,
            )

    def set_value_from_widget(self, widget, attr, value):
        """For subclasses."""
        ...

    @property
    def value_from_widget(self):
        return getattr(self._property_widget, self._property_name, None)

    @property
    def saved_value(self) -> Optional[Any]:
        raise None

    def save_settings(self):
        value = self.saved_value
        if value is not None:
            update_property_for_widget(
                self._property_widget,
                self._property_name,
                value
            )


class PropertyRuleEditor(_PropertyHelper, QtWidgets.QPushButton):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setAutoDefault(False)
        self.setDefault(False)
        self.clicked.connect(self._open_rules_editor)
        self.setText("&Rules...")

    def _open_rules_editor(self):
        from .rules_editor import RulesEditor
        self._rules_editor = RulesEditor(self._property_widget, parent=self)
        self._rules_editor.exec_()

    @property
    def saved_value(self) -> Optional[str]:
        return None


class PropertyCheckbox(_PropertyHelper, QtWidgets.QCheckBox):
    def set_value_from_widget(self, widget, attr, value):
        self.setChecked(bool(value))

    @property
    def saved_value(self) -> bool:
        return self.isChecked()


class PropertyLineEdit(_PropertyHelper, QtWidgets.QLineEdit):
    def set_value_from_widget(self, widget, attr, value):
        self.setText(value or "")

    @property
    def saved_value(self) -> Optional[str]:
        return self.text().strip()


class PropertyIntSpinBox(_PropertyHelper, QtWidgets.QSpinBox):
    def set_value_from_widget(self, widget, attr, value):
        self.setValue(value)

    @property
    def saved_value(self) -> int:
        return self.value()


class PropertyMacroTable(_PropertyHelper, DictionaryTable):
    def set_value_from_widget(self, widget, attr, value):
        try:
            macros = parse_macro_string(value or "")
        except Exception:
            logger.exception("Failed to parse macro string: %r", value)
        else:
            self.dictionary = macros

    @property
    def saved_value(self) -> Optional[str]:
        return json.dumps(self.dictionary)


class PropertyStringList(_PropertyHelper, StringListTable):
    def set_value_from_widget(self, widget, attr, value):
        self.values = value

    @property
    def saved_value(self) -> Optional[List[str]]:
        return self.values


def get_qt_properties(cls):
    """Yields all QMetaProperty instances from a given class."""
    meta_obj = cls.staticMetaObject
    for prop_idx in range(meta_obj.propertyCount()):
        prop = meta_obj.property(prop_idx)
        if prop is not None and prop.isDesignable():
            yield prop.name()


def get_helper_label_text(attr: str) -> str:
    spaced = re.sub("(PyDM|.)([A-Z])", r"\1 \2", attr)
    return spaced.strip().capitalize()


class BasicSettingsEditor(QtWidgets.QDialog):
    """
    QDialog for user-friendly editing of essential PyDM properties in Designer.

    Parameters
    ----------
    widget : PyDMWidget
        The widget which we want to edit.
    """

    _common_attributes_ = {
        "channel": PropertyLineEdit,
        "display": PropertyLineEdit,
        "macros": PropertyMacroTable,
        "filenames": PropertyStringList,
        "rules": PropertyRuleEditor,
    }

    _type_to_widget_ = {
        str: PropertyLineEdit,
        int: PropertyIntSpinBox,
        bool: PropertyCheckbox,
        "QStringList": PropertyStringList,
    }

    def __init__(self, widget, parent=None):
        super(BasicSettingsEditor, self).__init__(parent)

        self.widget = widget

        # PV names can be pretty wide...
        self.setMinimumSize(400, 150)

        self.setSizePolicy(
            QtWidgets.QSizePolicy.MinimumExpanding,
            QtWidgets.QSizePolicy.MinimumExpanding,
        )

        self.property_widgets = []
        self.setup_ui()

    def setup_ui(self):
        """
        Create the required UI elements for the form.

        Returns
        -------
        None
        """
        self.setWindowTitle("PyDM Widget Basic Settings Editor")
        vlayout = QtWidgets.QVBoxLayout()
        vlayout.setContentsMargins(5, 5, 5, 5)
        vlayout.setSpacing(5)
        self.setLayout(vlayout)

        settings_form = QtWidgets.QFormLayout()
        settings_form.setFieldGrowthPolicy(QtWidgets.QFormLayout.ExpandingFieldsGrow)
        vlayout.addLayout(settings_form)

        for helper_widget in self._create_helper_widgets(settings_form):
            self.property_widgets.append(helper_widget)

        buttons_layout = QtWidgets.QHBoxLayout()
        self.save_btn = QtWidgets.QPushButton("&Save", parent=self)
        self.save_btn.setAutoDefault(True)
        self.save_btn.setDefault(True)
        self.save_btn.clicked.connect(self.save_changes)
        self.update_btn = QtWidgets.QPushButton("&Update", parent=self)
        self.update_btn.clicked.connect(self.save_changes)
        cancel_btn = QtWidgets.QPushButton("&Cancel", parent=self)
        cancel_btn.clicked.connect(self.cancel_changes)
        buttons_layout.addStretch()
        buttons_layout.addWidget(cancel_btn)
        buttons_layout.addWidget(self.update_btn)
        buttons_layout.addWidget(self.save_btn)

        vlayout.addLayout(buttons_layout)

    def _create_helper_widgets(self, settings_form: QtWidgets.QFormLayout):
        other_attrs = []
        for attr in sorted(get_qt_properties(type(self.widget))):
            if attr not in self._common_attributes_ and attr not in other_attrs:
                other_attrs.append(attr)

        for attr in list(self._common_attributes_) + other_attrs:
            prop = getattr(type(self.widget), attr, None)
            if prop is None:
                continue

            prop_type = getattr(prop, "type", None)
            helper_widget_cls = self._common_attributes_.get(
                attr,
                self._type_to_widget_.get(prop_type, None)
            )
            if helper_widget_cls is not None:
                helper_widget = helper_widget_cls(
                    property_widget=self.widget,
                    property_name=attr,
                )
                label_text = get_helper_label_text(attr)
                settings_form.addRow(f"&{label_text}", helper_widget)
                yield helper_widget

    @QtCore.Slot()
    def save_changes(self):
        """Save the new settings on the widget properties."""
        for helper in self.property_widgets:
            try:
                helper.save_settings()
            except Exception:
                logger.exception(
                    "Failed to save settings for %s.%s = %r",
                    self.widget.objectName(),
                    helper._property_name,
                    helper.saved_value
                )

        if self.sender() == self.save_btn:
            self.accept()

    @QtCore.Slot()
    def cancel_changes(self):
        """Abort the changes and close the dialog."""
        self.close()



from qtpy.QtWidgets import QWidget, QTabWidget, QGridLayout, QLabel, QStyle, QStyleOption
from qtpy.QtGui import QColor, QPen, QFontMetrics, QPainter, QPaintEvent, QBrush
from qtpy.QtCore import Property, Qt, QSize, QPoint
from typing import List, Optional
from .base import PyDMWidget


class PyDMBitIndicator(QWidget):
    """
    A QWidget which draws a colored circle or rectangle

    Parameters
    ----------
    parent : QWidget
        The parent widget for the Label

    """
    def __init__(self, parent: Optional[QWidget] = None, circle: bool = False):
        super(PyDMBitIndicator, self).__init__(parent)
        self.circle = circle
        self._painter = QPainter()
        self._brush = QBrush(Qt.SolidPattern)
        self._pen = QPen(Qt.SolidLine)

    def paintEvent(self, event: QPaintEvent) -> None:
        """
        Paint events are sent to widgets that need to update themselves,
        for instance when part of a widget is exposed because a covering
        widget was moved.

        Parameters
        ----------
        event : QPaintEvent
        """
        self._painter.begin(self)
        opt = QStyleOption()
        opt.initFrom(self)
        self.style().drawPrimitive(QStyle.PE_Widget, opt, self._painter, self)
        self._painter.setRenderHint(QPainter.Antialiasing)
        self._painter.setBrush(self._brush)
        self._painter.setPen(self._pen)
        if self.circle:
            rect = self.rect()
            w = rect.width()
            h = rect.height()
            r = int(min(w, h) / 2.0 - 2.0 * max(self._pen.widthF(), 1.0))
            self._painter.drawEllipse(QPoint(w // 2, h // 2), r, r)
        else:
            self._painter.drawRect(self.rect())
        self._painter.end()

    def setColor(self, color: QColor) -> None:
        """
        Property for the color to be used when drawing

        Parameters
        ----------
        QColor
        """
        self._brush.setColor(color)
        self.update()

    def minimumSizeHint(self) -> QSize:
        fm = QFontMetrics(self.font())
        return QSize(fm.height(), fm.height())


class PyDMByteIndicator(QWidget, PyDMWidget):
    """
    Widget for graphical representation of bits from an integer number
    with support for Channels and more from PyDM

    Parameters
    ----------
    parent : QWidget
        The parent widget for the Label
    init_channel : str, optional
        The channel to be used by the widget.
    """
    def __init__(self, parent: Optional[QWidget] = None, init_channel=None):
        QWidget.__init__(self, parent)
        PyDMWidget.__init__(self, init_channel=init_channel)
        self.value = 0
        self.setLayout(QGridLayout(self))

        self._on_color = QColor(0, 255, 0)
        self._off_color = QColor(100, 100, 100)
        self._disconnected_color = QColor(255, 255, 255)
        self._invalid_color = QColor(255, 0, 255)

        self._pen_style = Qt.SolidLine
        self._line_pen = QPen(self._pen_style)

        self._orientation = Qt.Vertical

        # This is kind of ridiculous, importing QTabWidget just to get a 4-item enum thats usable in Designer.
        # PyQt5 lets you define custom enums that you can use in designer with QtCore.Q_ENUMS(), doesn't exist in PyQt4.
        self._labels = []
        self._show_labels = True
        self._label_position = QTabWidget.East

        self._num_bits = 1

        self._indicators = []
        self._circles = False
        self.set_spacing()
        self.layout().setOriginCorner(Qt.TopLeftCorner)

        self._big_endian = False
        self._shift = 0
        self.numBits = 1  # Need to set the property to initialize
        # _labels and _indicators setting numBits there also performs
        # the first rebuild_layout.

    def init_for_designer(self) -> None:
        """
        Method called after the constructor to tweak configurations for
        when using the widget with the Qt Designer
        """
        self._connected = True
        self.value = 5
        self.update_indicators()

    def connection_changed(self, connected: bool) -> None:
        """
        Callback invoked when the connection state of the Channel is changed.
        This callback acts on the connection state to enable/disable the widget
        and also trigger the change on alarm severity to ALARM_DISCONNECTED.

        Parameters
        ----------
        connected : bool
            When this value is False the channel is disconnected, True otherwise.
        """
        super(PyDMByteIndicator, self).connection_changed(connected)
        self.update_indicators()

    def rebuild_layout(self) -> None:
        """
        Method to reorganize the top-level widget and its contents
        according to the layout property values.
        """
        self.clear()
        pairs = zip(self._labels, self._indicators)
        # Hide labels until they are in the layout
        for label in self._labels:
            label.setVisible(False)
        # This is a horrendous mess of if statements
        # for every possible case.  Ugh.
        # There is probably a more clever way to do this.
        if self.orientation == Qt.Vertical:
            for i, (label, indicator) in enumerate(pairs):
                if self.labelPosition == QTabWidget.East:
                    self.layout().addWidget(indicator, i, 0)
                    self.layout().addWidget(label, i, 1, 1, 1, Qt.AlignVCenter)
                    label.setVisible(self._show_labels)
                elif self.labelPosition == QTabWidget.West:
                    self.layout().addWidget(label, i, 0, 1, 1, Qt.AlignVCenter)
                    self.layout().addWidget(indicator, i, 1)
                    label.setVisible(self._show_labels)
                else:
                    self.layout().addWidget(indicator, i, 0)
                    # Invalid combo of orientation and label position,
                    # so we don't reset label visibility here.
        elif self.orientation == Qt.Horizontal:
            for i, (label, indicator) in enumerate(pairs):
                if self.labelPosition == QTabWidget.North:
                    self.layout().addWidget(label, 0, i, 1, 1, Qt.AlignHCenter)
                    self.layout().addWidget(indicator, 1, i)
                    label.setVisible(self._show_labels)
                elif self.labelPosition == QTabWidget.South:
                    self.layout().addWidget(indicator, 0, i)
                    self.layout().addWidget(label, 1, i, 1, 1, Qt.AlignHCenter)
                    label.setVisible(self._show_labels)
                else:
                    self.layout().addWidget(indicator, 0, i)
                    # Invalid combo of orientation and label position, so
                    # we don't reset label visibility here.
        self.update_indicators()

    def clear(self) -> None:
        """
        Remove all inner widgets from the layout
        """
        for col in range(0, self.layout().columnCount()):
            for row in range(0, self.layout().rowCount()):
                item = self.layout().itemAtPosition(row, col)
                if item is not None:
                    w = item.widget()
                    if w is not None:
                        self.layout().removeWidget(w)

    def update_indicators(self) -> None:
        """
        Update the inner bit indicators accordingly with the new value.
        """
        if self._shift < 0:
            value = int(self.value) << abs(self._shift)
        else:
            value = int(self.value) >> self._shift
        if value < 0:
            value = 0

        bits = [(value >> i) & 1
                for i in range(self._num_bits)]
        for bit, indicator in zip(bits, self._indicators):
            if self._connected:
                if self._alarm_state == 3:
                    c = self._invalid_color
                else:
                    c = self._on_color if bit else self._off_color
            else:
                c = self._disconnected_color
            indicator.setColor(c)

    @Property(QColor)
    def onColor(self) -> QColor:
        """
        The color for a bit in the 'on' state.

        Returns
        -------
        QColor
        """
        return self._on_color

    @onColor.setter
    def onColor(self, new_color: QColor) -> None:
        """
        The color for a bit in the 'on' state.

        Parameters
        ----------
        new_color : QColor
        """
        if new_color != self._on_color:
            self._on_color = new_color
            self.update_indicators()

    @Property(QColor)
    def offColor(self) -> QColor:
        """
        The color for a bit in the 'off' state.

        Returns
        -------
        QColor
        """
        return self._off_color

    @offColor.setter
    def offColor(self, new_color: QColor) -> None:
        """
        The color for a bit in the 'off' state.

        Parameters
        ----------
        new_color : QColor
        """
        if new_color != self._off_color:
            self._off_color = new_color
            self.update_indicators()

    @Property(Qt.Orientation)
    def orientation(self) -> Qt.Orientation:
        """
        Whether to lay out the bit indicators vertically or horizontally.

        Returns
        -------
        int
        """
        return self._orientation

    @orientation.setter
    def orientation(self, new_orientation: Qt.Orientation) -> None:
        """
        Whether to lay out the bit indicators vertically or horizontally.

        Parameters
        -------
        new_orientation : Qt.Orientation, int
        """
        self._orientation = new_orientation
        self.set_spacing()
        self.rebuild_layout()

    def set_spacing(self) -> None:
        """
        Configures the correct spacing given the selected orientation.
        """
        label_spacing = 5
        if self._circles:
            indicator_spacing = 5
        else:
            indicator_spacing = 0

        self.layout().setContentsMargins(0, 0, 0, 0)

        if self._orientation == Qt.Horizontal:
            self.layout().setHorizontalSpacing(indicator_spacing)
            self.layout().setVerticalSpacing(label_spacing)
        elif self._orientation == Qt.Vertical:
            self.layout().setHorizontalSpacing(label_spacing)
            self.layout().setVerticalSpacing(indicator_spacing)

    @Property(bool)
    def showLabels(self) -> bool:
        """
        Whether or not to show labels next to each bit indicator.

        Returns
        -------
        bool
        """
        return self._show_labels

    @showLabels.setter
    def showLabels(self, show: bool) -> None:
        """
        Whether or not to show labels next to each bit indicator.

        Parameters
        -------
        show : bool
            If True the widget will show a label next to the bit indicator
        """
        self._show_labels = show
        for label in self._labels:
            label.setVisible(show)

    @Property(bool)
    def bigEndian(self) -> bool:
        """
        Whether the most significant bit is at the start or end of the widget.

        Returns
        -------
        bool
        """
        return self._big_endian

    @bigEndian.setter
    def bigEndian(self, is_big_endian: bool) -> None:
        """
        Whether the most significant bit is at the start or end of the widget.

        Parameters
        -------
        is_big_endian : bool
            If True, the Big Endian will be used, Little Endian otherwise
        """
        self._big_endian = is_big_endian

        origin_map = {(Qt.Vertical, True): Qt.BottomLeftCorner,
                      (Qt.Vertical, False): Qt.TopLeftCorner,
                      (Qt.Horizontal, True): Qt.TopRightCorner,
                      (Qt.Horizontal, False): Qt.TopLeftCorner}

        origin = origin_map[(self.orientation, self.bigEndian)]
        self.layout().setOriginCorner(origin)
        self.rebuild_layout()

    @Property(bool)
    def circles(self) -> bool:
        """
        Draw indicators as circles, rather than rectangles.

        Returns
        -------
        bool
        """
        return self._circles

    @circles.setter
    def circles(self, draw_circles: bool) -> None:
        """
        Draw indicators as circles, rather than rectangles.

        Parameters
        ----------
        draw_circles : bool
            If True, bits will be represented as circles
        """
        self._circles = draw_circles
        self.set_spacing()
        for indicator in self._indicators:
            indicator.circle = self._circles
        self.update_indicators()

    @Property(QTabWidget.TabPosition)
    def labelPosition(self) -> QTabWidget.TabPosition:
        """
        The side of the widget to display labels on.

        Returns
        -------
        int
        """
        return self._label_position

    @labelPosition.setter
    def labelPosition(self, new_pos: QTabWidget.TabPosition) -> None:
        """
        The side of the widget to display labels on.

        Parameters
        ----------
        new_pos : QTabWidget.TabPosition, int
        """
        self._label_position = new_pos
        self.rebuild_layout()

    @Property(int)
    def numBits(self) -> int:
        """
        Number of bits to interpret.

        Returns
        -------
        int
        """
        return self._num_bits

    @numBits.setter
    def numBits(self, new_num_bits: int) -> None:
        """
        Number of bits to interpret.

        Parameters
        ----------
        new_num_bits : int
        """
        if new_num_bits < 1:
            return
        self._num_bits = new_num_bits
        for indicator in self._indicators:
            indicator.deleteLater()
        self._indicators = [PyDMBitIndicator(parent=self, circle=self.circles)
                            for i in range(0, self._num_bits)]
        old_labels = self.labels
        new_labels = ["Bit {}".format(i) for i in range(0, self._num_bits)]
        for i, old_label in enumerate(old_labels):
            if i >= self._num_bits:
                break
            new_labels[i] = old_label
        self.labels = new_labels

    @Property(int)
    def shift(self) -> int:
        """
        Bit shift.

        Returns
        -------
        int
        """
        return self._shift

    @shift.setter
    def shift(self, new_shift: int) -> None:
        """
        Bit shift.

        Parameters
        ----------
        new_shift : int
        """
        self._shift = new_shift
        self.update_indicators()

    @Property('QStringList')
    def labels(self) -> List[str]:
        """
        Labels for each bit.

        Returns
        -------
        list
        """
        return [str(l.text()) for l in self._labels]

    @labels.setter
    def labels(self, new_labels: List[str]) -> None:
        """
        Labels for each bit.

        Parameters
        ----------
        new_labels : list
        """
        for label in self._labels:
            label.hide()
            label.deleteLater()
        self._labels = [QLabel(text, parent=self) for text in new_labels]
        # Have to reset showLabels to hide or show all the new labels we just made.
        self.showLabels = self._show_labels
        self.rebuild_layout()

    def value_changed(self, new_val: int) -> None:
        """
        Callback invoked when the Channel value is changed.

        Parameters
        ----------
        new_val : int
            The new value from the channel.
        """
        super(PyDMByteIndicator, self).value_changed(new_val)
        try:
            int(new_val)
            self.update_indicators()
        except:
            pass

    def paintEvent(self, _) -> None:
        """
        Paint events are sent to widgets that need to update themselves,
        for instance when part of a widget is exposed because a covering
        widget was moved.

        At PyDMDrawing this method handles the alarm painting with parameters
        from the stylesheet, configures the brush, pen and calls ```draw_item```
        so the specifics can be performed for each of the drawing classes.

        Parameters
        ----------
        event : QPaintEvent
        """
        painter = QPainter(self)
        opt = QStyleOption()
        opt.initFrom(self)
        self.style().drawPrimitive(QStyle.PE_Widget, opt, painter, self)
        painter.setRenderHint(QPainter.Antialiasing)

    def alarm_severity_changed(self, new_alarm_severity: int) -> None:
        """
        Callback invoked when the Channel alarm severity is changed.

        Parameters
        ----------
        new_alarm_severity : int
            The new severity where 0 = NO_ALARM, 1 = MINOR, 2 = MAJOR
            and 3 = INVALID
        """

        if new_alarm_severity == self._alarm_state:
            return
        else:
            super(PyDMByteIndicator, self).alarm_severity_changed(new_alarm_severity)

            # Checks if _shift attribute exits because base class can call method
            # before the object constructor is complete
            if hasattr(self, '_shift'):
                self.update_indicators()



# Unit Tests for the Frame Widget


import pytest

from ...widgets.base import is_channel_valid
from ... import data_plugins
from ...widgets.base import PyDMWidget
from ...widgets.frame import PyDMFrame


# --------------------
# POSITIVE TEST CASES
# --------------------

def test_construct(qtbot):
    """
    Test the construction of the widget.

    Expectations:
    The correct default values are assigned to the widget's attributes.

    Parameters
    ----------
    qtbot : fixture
        pytest-qt window for widget test
    """
    pydm_frame = PyDMFrame()
    qtbot.addWidget(pydm_frame)

    assert pydm_frame._disable_on_disconnect is False
    assert pydm_frame.alarmSensitiveBorder is False


@pytest.mark.parametrize("init_value, new_value", [
    (False, True),
    (True, False),
    (False, False),
    (True, True)
])
def test_disable_on_disconnect(qtbot, init_value, new_value):
    """
    Test setting the flag to disable the widget when there's a channel disconnection.

    Expectations:
    The widget retains the new Boolean setting.

    Parameters
    ----------
    qtbot : fixture
        pytest-qt window for widget test
    init_value : bool
        The initial flag to whether disable a widget at disconnection. True is to disable the widget in that event;
        False is not to disable.
    new_value
        The flag to change the widget disabling setting. True is to disable the widget in that event;
        False is not to disable.
    """
    pydm_frame = PyDMFrame()
    qtbot.addWidget(pydm_frame)

    pydm_frame._disable_on_disconnect = init_value
    pydm_frame.disableOnDisconnect = new_value
    assert pydm_frame.disableOnDisconnect == new_value


@pytest.mark.parametrize("channel, alarm_sensitive_content, alarm_sensitive_border, new_alarm_severity", [
    (None, False, False, PyDMWidget.ALARM_NONE),
    (None, False, True, PyDMWidget.ALARM_NONE),
    (None, True, False, PyDMWidget.ALARM_NONE),
    (None, True, True, PyDMWidget.ALARM_NONE),

    (None, False, False, PyDMWidget.ALARM_MAJOR),
    (None, False, True, PyDMWidget.ALARM_MAJOR),
    (None, True, False, PyDMWidget.ALARM_MAJOR),
    (None, True, True, PyDMWidget.ALARM_MAJOR),

    ("CA://MTEST", False, False, PyDMWidget.ALARM_NONE),
    ("CA://MTEST", False, True, PyDMWidget.ALARM_NONE),
    ("CA://MTEST", True, False, PyDMWidget.ALARM_NONE),
    ("CA://MTEST", True, True, PyDMWidget.ALARM_NONE),

    ("CA://MTEST", False, False, PyDMWidget.ALARM_MINOR),
    ("CA://MTEST", False, True, PyDMWidget.ALARM_MINOR),
    ("CA://MTEST", True, False, PyDMWidget.ALARM_MINOR),
    ("CA://MTEST", True, True, PyDMWidget.ALARM_MINOR),

    ("CA://MTEST", False, False, PyDMWidget.ALARM_MAJOR),
    ("CA://MTEST", False, True, PyDMWidget.ALARM_MAJOR),
    ("CA://MTEST", True, False, PyDMWidget.ALARM_MAJOR),
    ("CA://MTEST", True, True, PyDMWidget.ALARM_MAJOR),

    ("CA://MTEST", False, False, PyDMWidget.ALARM_DISCONNECTED),
    ("CA://MTEST", False, True, PyDMWidget.ALARM_DISCONNECTED),
    ("CA://MTEST", True, False, PyDMWidget.ALARM_DISCONNECTED),
    ("CA://MTEST", True, True, PyDMWidget.ALARM_DISCONNECTED),
])
def test_alarm_severity_change(qtbot, signals, channel, alarm_sensitive_content, alarm_sensitive_border,
                               new_alarm_severity):
    """
    Test the style of the widget changing according to alarm sensitivity settings and alarm severity changes.

    Expectations:
    Depending on the initial widget's settings on whether the widget should change its content area and borders when
    there's an alarm event, the widget's style should reflect changes when there's an alarm event other than ALARM_NONE.

    Parameters
    ----------
    qtbot : fixture
        pytest-qt window for widget test
    signals : fixture
        The signals fixture, which provides access signals to be bound to the appropriate slots
    channel : str
        The data channel address
    alarm_sensitive_content : bool
        True if the content area of the widget will change its color when an alarm happens; False if not
    alarm_sensitive_border : bool
        True if the borders of the widget will change its color when an alarm happens; False if not
    new_alarm_severity : PyDMWidget alarm type
        The new alarm severity that may prompt the widget to change its content area and/or border colors.
    """
    pydm_frame = PyDMFrame()
    qtbot.addWidget(pydm_frame)

    pydm_frame._channel = channel
    pydm_frame.alarmSensitiveContent = alarm_sensitive_content
    pydm_frame.alarmSensitiveBorder = alarm_sensitive_border


@pytest.mark.parametrize("channel_address, connected, write_access, is_app_read_only", [
    ("CA://MA_TEST", True, True, True),
    ("CA://MA_TEST", True, False, True),
    ("CA://MA_TEST", True, True, False),
    ("CA://MA_TEST", True, False, False),
    ("CA://MA_TEST", False, True, True),
    ("CA://MA_TEST", False, False, True),
    ("CA://MA_TEST", False, True, False),
    ("CA://MA_TEST", False, False, False),
    ("", False, False, False),
    (None, False, False, False),
])
def test_check_enable_state(qtbot, signals, channel_address, connected, write_access, is_app_read_only):
    """
    Test the tooltip generated depending on the channel address validation, connection, write access, and whether the
    app is read-only.

    Expectations:
    1. The widget's tooltip will update only if the channel address is valid.
    2. If the data channel is disconnected, the widget's tooltip will  "PV is disconnected"
    3. If the data channel is connected, but it has no write access:
        a. If the app is read-only, the tooltip will read  "Running PyDM on Read-Only mode."
        b. If the app is not read-only, the tooltip will read "Access denied by Channel Access Security."

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    signals : fixture
        The signals fixture, which provides access signals to be bound to the appropriate slots
    channel_address : str
        The channel address
    connected : bool
        True if the channel is connected; False otherwise
    write_access : bool
        True if the widget has write access to the channel; False otherwise
    is_app_read_only : bool
        True if the PyDM app is read-only; False otherwise
    """
    pydm_frame = PyDMFrame()
    qtbot.addWidget(pydm_frame)

    pydm_frame.channel = channel_address

    if not pydm_frame._disable_on_disconnect:
        assert pydm_frame.isEnabled()
    else:
        signals.write_access_signal[bool].connect(pydm_frame.writeAccessChanged)
        signals.write_access_signal[bool].emit(write_access)

        signals.connection_state_signal[bool].connect(pydm_frame.connectionStateChanged)
        signals.connection_state_signal[bool].emit(connected)

        data_plugins.set_read_only(is_app_read_only)

        original_tooltip = "Original Tooltip"
        pydm_frame.setToolTip(original_tooltip)
        pydm_frame.check_enable_state()

        actual_tooltip = pydm_frame.toolTip()
        if is_channel_valid(channel_address):
            if not pydm_frame._connected:
                assert "PV is disconnected." in actual_tooltip
            elif not write_access:
                if data_plugins.is_read_only():
                    assert "Running PyDM on Read-Only mode." in actual_tooltip
                else:
                    assert "Access denied by Channel Access Security." in actual_tooltip
        else:
            assert actual_tooltip == original_tooltip



import os
from os import path

from qtpy.QtWidgets import (QApplication, QMainWindow, QFileDialog,
                            QAction, QMessageBox)
from qtpy.QtCore import Qt, QTimer, Slot, QSize, QLibraryInfo
from qtpy.QtGui import QKeySequence
from .utilities import (IconFont, find_file, establish_widget_connections,
                        close_widget_connections)
from .pydm_ui import Ui_MainWindow
from .display import Display, ScreenTarget, load_file
from .connection_inspector import ConnectionInspector
from .about_pydm import AboutWindow
from .show_macros import MacroWindow
from . import data_plugins
from . import tools
from .widgets.rules import register_widget_rules, unregister_widget_rules
from . import config
import subprocess
import platform
import logging

from .utilities.stylesheet import apply_stylesheet, _get_style_data

logger = logging.getLogger(__name__)


class PyDMMainWindow(QMainWindow):

    def __init__(self, parent=None, hide_nav_bar=False, hide_menu_bar=False, hide_status_bar=False,
                 home_file=None, macros=None, command_line_args=None):
        super(PyDMMainWindow, self).__init__(parent)
        self.app = QApplication.instance()
        self.font_factor = 1
        self.iconFont = IconFont()
        self._display_widget = None
        self._showing_file_path_in_title_bar = False

        # style sheet change flag
        self.isSS_Changed = False
        self.ss_path = None

        self._saved_menu_geometry = None
        self._saved_menu_height = None
        self._new_widget_size = None

        self.default_font_size = QApplication.instance().font().pointSizeF()

        self.home_file = home_file
        self.home_widget = None
        if home_file:
            self.home_widget = load_file(home_file, macros=macros, args=command_line_args, target=ScreenTarget.HOME)
            close_widget_connections(self.home_widget)

        self.designer_path = None

        self.ui = Ui_MainWindow()
        self.ui.setupUi(self)

        self.showMacros = QAction("Show Macros...", self)

        self.ui.navbar.setIconSize(QSize(24, 24))
        self.ui.navbar.setToolButtonStyle(Qt.ToolButtonTextUnderIcon)
        self.ui.actionHome.triggered.connect(self.home_triggered)
        self.ui.actionHome.setIcon(self.iconFont.icon("home"))
        self.ui.actionBack.triggered.connect(self.back_triggered)
        self.ui.actionBack.setIcon(self.iconFont.icon("angle-left"))
        self.ui.actionForward.triggered.connect(self.forward_triggered)
        self.ui.actionForward.setIcon(self.iconFont.icon("angle-right"))
        self.ui.actionEdit_in_Designer.triggered.connect(self.edit_in_designer)
        self.ui.actionOpen_File.triggered.connect(self.open_file_action)
        self.ui.actionChange_Stylesheet.triggered.connect(self.change_stylesheet_action)
        self.ui.actionReload_Display.triggered.connect(self.reload_display)
        self.ui.actionIncrease_Font_Size.triggered.connect(self.increase_font_size)
        self.ui.actionDecrease_Font_Size.triggered.connect(self.decrease_font_size)
        self.ui.actionDefault_Font_Size.triggered.connect(self.reset_font_size)
        self.ui.actionEnter_Fullscreen.triggered.connect(self.enter_fullscreen)
        self.ui.actionShow_File_Path_in_Title_Bar.triggered.connect(self.toggle_file_path_in_title_bar)
        self.ui.actionShow_Navigation_Bar.triggered.connect(self.toggle_nav_bar)
        self.ui.actionShow_Menu_Bar.triggered.connect(self.toggle_menu_bar)
        self.ui.actionShow_Status_Bar.triggered.connect(self.toggle_status_bar)
        self.ui.actionShow_Connections.triggered.connect(self.show_connections)
        self.ui.actionShow_Help.triggered.connect(self.show_help)
        self.ui.actionAbout_PyDM.triggered.connect(self.show_about_window)
        self.ui.actionLoadTool.triggered.connect(self.load_tool)
        self.ui.actionLoadTool.setIcon(self.iconFont.icon("rocket"))
        self.showMacros.triggered.connect(self.show_macro_window)
        self.ui.actionQuit.triggered.connect(self.quit_main_window)

        if hide_nav_bar:
            self.toggle_nav_bar(False)
            self.ui.actionShow_Navigation_Bar.setChecked(False)
        if hide_menu_bar:
            # Toggle the menu bar via the QAction so that the menu item
            # stays in sync with menu visibility.
            self.ui.actionShow_Menu_Bar.activate(QAction.Trigger)
        if hide_status_bar:
            self.toggle_status_bar(False)
            self.ui.actionShow_Status_Bar.setChecked(False)
        # Try to find the designer binary.
        self.ui.actionEdit_in_Designer.setEnabled(False)

        possible_designer_bin_paths = (QLibraryInfo.location(QLibraryInfo.BinariesPath), QLibraryInfo.location(QLibraryInfo.LibraryExecutablesPath))
        for bin_path in possible_designer_bin_paths:
            if platform.system() == 'Darwin':
                designer_path = os.path.join(bin_path, 'Designer.app/Contents/MacOS/Designer')
            elif platform.system() == 'Linux':
                designer_path = os.path.join(bin_path, 'designer')
            else:
                designer_path = os.path.join(bin_path, 'designer.exe')
            if os.path.isfile(designer_path):
                self.designer_path = designer_path
                break

        self.update_tools_menu()
        self.enable_disable_navigation()

    def display_widget(self):
        return self._display_widget

    def set_display_widget(self, new_widget):
        if new_widget == self._display_widget:
            return
        self.clear_display_widget()
        if not new_widget.layout():
            new_widget.setMinimumSize(new_widget.size())
        self._new_widget_size = new_widget.size()
        new_widget.setVisible(True)
        self._display_widget = new_widget
        self.setCentralWidget(self._display_widget)
        self.enable_disable_navigation()
        self.update_window_title()
        self.add_menu_items()
        # Resizing to the new widget's dimensions needs to be
        # done on the event loop for some reason - you can't
        # just do it here.
        QTimer.singleShot(0, self.resizeForNewDisplayWidget)

    def clear_display_widget(self):
        if self._display_widget is not None:
            close_widget_connections(self._display_widget)
            unregister_widget_rules(self._display_widget)
            self._display_widget.setVisible(False)
            self._display_widget.setParent(None)
            self.ui.actionEdit_in_Designer.setEnabled(False)

    def handle_open_file_error(self, filename, error):
        self.statusBar().showMessage("Cannot open file: '{0}', reason: '{1}'...".format(filename, error), 5000)

    def back_triggered(self):
        new_process = QApplication.keyboardModifiers() == Qt.ShiftModifier
        self.back(open_in_new_process=new_process)

    def back(self, open_in_new_process=False):
        curr_display = self.display_widget()
        prev_display = None
        if curr_display:
            prev_display = curr_display.previous_display

        if not prev_display:
            logger.error('No display history to execute back navigation.')
            return

        if open_in_new_process:
            load_file(prev_display.loaded_file(),
                      macros=prev_display.macros,
                      args=prev_display.args,
                      target=ScreenTarget.NEW_PROCESS)
        else:
            prev_display.next_display = curr_display
            establish_widget_connections(prev_display)
            register_widget_rules(prev_display)
            self.set_display_widget(prev_display)

    def forward_triggered(self):
        new_process = QApplication.keyboardModifiers() == Qt.ShiftModifier
        self.forward(open_in_new_process=new_process)

    def forward(self, open_in_new_process=False):
        curr_display = self.display_widget()
        next_display = None
        if curr_display:
            next_display = curr_display.next_display

        if not next_display:
            logger.error('No display history to execute forward navigation.')
            return

        if open_in_new_process:
            load_file(next_display.loaded_file(),
                      macros=next_display.macros,
                      args=next_display.args,
                      target=ScreenTarget.NEW_PROCESS)
        else:
            establish_widget_connections(next_display)
            register_widget_rules(next_display)
            next_display.previous_display = curr_display
            self.set_display_widget(next_display)

    def home_triggered(self):
        new_process = QApplication.keyboardModifiers() == Qt.ShiftModifier
        self.home(open_in_new_process=new_process)

    def home(self, open_in_new_process=False):
        if self.home_widget is None:
            return

        if open_in_new_process:
            fname = self.home_widget.loaded_file()
            macros = self.home_widget.macros()
            args = self.home_widget.args()
            load_file(fname,
                      macros=macros,
                      args=args,
                      target=ScreenTarget.NEW_PROCESS)
        else:
            if self.home_widget != self.display_widget():
                self.home_widget.previous_display = self.display_widget()
            establish_widget_connections(self.home_widget)
            register_widget_rules(self.home_widget)
            self.set_display_widget(self.home_widget)

    def enable_disable_navigation(self):
        w = self.display_widget()
        if not isinstance(w, Display):
            # We can't do much if it is not a Display and we don't have the
            # previous_display and next_display properties since we don't
            # have the navigation stack set.
            nav_stack_methods = hasattr(w, 'previous_display') \
                                and hasattr(w, 'next_display')
            if not nav_stack_methods:
                return
        if not w:
            self.ui.actionBack.setDisabled(True)
            self.ui.actionForward.setDisabled(True)
            return
        self.ui.actionBack.setDisabled(w.previous_display is None)
        self.ui.actionForward.setDisabled(w.next_display is None)

    def current_file(self):
        return self.display_widget().loaded_file()

    def update_window_title(self):
        if self.showing_file_path_in_title_bar:
            title = self.current_file()
        else:
            title = self.display_widget().windowTitle()
        title += " - PyDM"
        if data_plugins.is_read_only():
            title += " [Read Only Mode]"
        self.setWindowTitle(title)

    @property
    def showing_file_path_in_title_bar(self):
        return self._showing_file_path_in_title_bar

    @showing_file_path_in_title_bar.setter
    def showing_file_path_in_title_bar(self, checked):
        self._showing_file_path_in_title_bar = checked
        self.update_window_title()

    @Slot(bool)
    def toggle_file_path_in_title_bar(self, checked):
        self.showing_file_path_in_title_bar = checked

    @Slot(bool)
    def toggle_nav_bar(self, checked):
        self.ui.navbar.setHidden(not checked)

    @Slot(bool)
    def toggle_menu_bar(self, checked=None):
        if checked is None:
            checked = not self.ui.menubar.isVisible()
        # Crazy hack: we can't just do menubar.setVisible(), because that
        # will disable all the QActions and their keyboard shortcuts when
        # we hide the menu.  So instead, we set it to a height of 0 to hide
        # it, and then restore the previous height value to show it again.
        if checked:
            self.ui.menubar.restoreGeometry(self._saved_menu_geometry)
            self.ui.menubar.setFixedHeight(self._saved_menu_height)
        else:
            self._saved_menu_geometry = self.ui.menubar.saveGeometry()
            self._saved_menu_height = self.ui.menubar.height()
            self.ui.menubar.setFixedHeight(0)

    @Slot(bool)
    def toggle_status_bar(self, checked):
        self.ui.statusbar.setHidden(not checked)

    def get_files_in_display(self):
        try:
            curr_file = self.current_file()
        except IndexError:
            logger.error("The display manager does not have a display loaded.")
            return None, None

        _, extension = path.splitext(curr_file)
        if extension == '.ui':
            return self.current_file(), None
        else:
            central_widget = self.centralWidget() if isinstance(self.centralWidget(), Display) else None
            if central_widget is not None:
                ui_file = central_widget.ui_filepath()
            return ui_file, self.current_file()

    @Slot(bool)
    def edit_in_designer(self, checked):

        def open_editor_ui(fname):
            if self.designer_path is None or fname is None or fname == "":
                return
            self.statusBar().showMessage("Launching '{0}' in Qt Designer...".format(fname), 5000)
            _ = subprocess.Popen('{0} "{1}"'.format(self.designer_path, fname), shell=True)

        def open_editor_generic(fname):
            if platform.system() == "Darwin":
                subprocess.call(('open', fname))
            elif platform.system() == "Linux":
                subprocess.call(('xdg-open', fname))
            elif platform.system() == "Windows":
                os.startfile(fname)

        ui_file, py_file = self.get_files_in_display()
        if ui_file is not None and ui_file != "":
            open_editor_ui(fname=ui_file)
        if py_file is not None and py_file != "":
            open_editor_generic(fname=py_file)

    @Slot(bool)
    def open_file_action(self, checked):
        modifiers = QApplication.keyboardModifiers()
        try:
            curr_file = self.current_file()
            folder = os.path.dirname(curr_file)
        except Exception:
            folder = os.getcwd()

        filename = QFileDialog.getOpenFileName(self, 'Open File...', folder, 'PyDM Display Files (*.ui *.py)')
        filename = filename[0] if isinstance(filename, (list, tuple)) else filename

        if filename:
            filename = str(filename)
            try:
                if modifiers == Qt.ShiftModifier:
                    target = ScreenTarget.NEW_PROCESS
                else:
                    target = None
                self.open(filename, target=target)

            except (IOError, OSError, ValueError, ImportError) as e:
                self.handle_open_file_error(filename, e)

    @Slot(bool)
    def change_stylesheet_action(self, checked):
        try:
            curr_file = self.current_file()
            folder = os.path.dirname(curr_file)
        except Exception:
            folder = os.getcwd()

        ss_filename = QFileDialog.getOpenFileName(self, 'Change Stylesheet...',
                folder, 'PyDM Stylesheets (*.qss *.css)')
        
        if ss_filename:
            ss_filename = str(ss_filename)
            self.ss_path = ss_filename.split("'")
            style= open(self.ss_path[1],"r")
            style= style.read()
            self.setStyleSheet(style)

    def open(self, filename, macros=None, args=None, target=None):
        if not os.path.isabs(filename):
            base_path = None
            curr_display = self.display_widget()
            if curr_display:
                base_path = os.path.dirname(curr_display.loaded_file())
            filename = find_file(filename, base_path=base_path, raise_if_not_found=True)
        new_widget = load_file(filename,
                               macros=macros,
                               args=args,
                               target=target)
        if new_widget:
            if self.home_widget is None:
                self.home_widget = new_widget
            display_widget = self.display_widget()
            if display_widget:
                new_widget.previous_display = display_widget
            self.set_display_widget(new_widget)
            ui_file, py_file = self.get_files_in_display()
            editors = []
            if ui_file:
                editors.append("Designer")
            if py_file:
                editors.append("Text Editor")
            edit_in_text = "Open in {}".format(' and '.join(editors))
            self.ui.actionEdit_in_Designer.setText(edit_in_text)
            if (self.designer_path and ui_file) or (py_file and not ui_file):
                self.ui.actionEdit_in_Designer.setEnabled(True)
        return new_widget

    def load_tool(self, checked):
        try:
            curr_dir = os.path.dirname(self.current_file())
        except Exception:
            curr_dir = os.getcwd()
            logger.error("The display manager does not have a display loaded. Suggesting current work directory.")
        filename = QFileDialog.getOpenFileName(self, 'Load tool...', curr_dir, 'PyDM External Tool Files (*_tool.py)')
        filename = filename[0] if isinstance(filename, (list, tuple)) else filename

        if filename:
            filename = str(filename)
            tools.install_external_tool(filename)
            self.update_tools_menu()

    def update_tools_menu(self):
        """
        Update the Main Window Tools menu.
        """
        kwargs = {'channels': None, 'sender': self}
        tools.assemble_tools_menu(self.ui.menuTools,
                                  clear_menu=True,
                                  **kwargs)

        self.ui.menuTools.addSeparator()
        self.ui.menuTools.addAction(self.ui.actionLoadTool)
        self.ui.menuTools.addAction(self.showMacros)

    @Slot(bool)
    def reload_display(self, checked):
        curr_display = self.display_widget()
        if not curr_display:
            logger.error("The display manager does not have a display loaded.")
            return

        prev_display = curr_display.previous_display
        next_display = curr_display.next_display

        macros = curr_display.macros()
        args = curr_display.args()
        loaded_file = curr_display.loaded_file()

        self.statusBar().showMessage(
            "Reloading '{0}'...".format(self.current_file()), 5000)
        new_widget = self.open(loaded_file, macros=macros, args=args)
        new_widget.previous_display = prev_display
        new_widget.next_display = next_display

    @Slot(bool)
    def increase_font_size(self, checked):
        old_factor = self.font_factor
        self.font_factor += 0.1
        self.set_font_size(old_factor, self.font_factor)

    @Slot(bool)
    def decrease_font_size(self, checked):
        old_factor = self.font_factor
        self.font_factor -= 0.1
        self.set_font_size(old_factor, self.font_factor)

    @Slot(bool)
    def reset_font_size(self, checked):
        old_factor = self.font_factor
        self.font_factor = 1
        self.set_font_size(old_factor, self.font_factor)

    def set_font_size(self, old, new):
        current_font = self.app.font()
        current_font.setPointSizeF(current_font.pointSizeF()/old*new)
        QApplication.instance().setFont(current_font)

        for w in self.app.allWidgets():
            w_c_f = w.font()
            w_c_f.setPointSizeF(w_c_f.pointSizeF()/old*new)
            w.setFont(w_c_f)

        QTimer.singleShot(0, self.resizeForNewDisplayWidget)

    @Slot(bool)
    def enter_fullscreen(self, checked=False):
        if self.isFullScreen():
            self.showNormal()
        else:
            self.showFullScreen()

    @Slot(bool)
    def show_connections(self, checked):
        c = ConnectionInspector(self)
        c.show()

    def show_help(self):
        """ Show the associated help file for this window """
        if self.display_widget() is not None:
            self.display_widget().show_help()

    @Slot(bool)
    def show_about_window(self, checked):
        a = AboutWindow(self)
        a.show()

    def resizeForNewDisplayWidget(self):
        if not self.isFullScreen():
            self.resize(self._new_widget_size)

    def closeEvent(self, event):
        event.ignore()

        def do_close():
            self.clear_display_widget()
            event.accept()

        main_windows = [w for w in self.app.topLevelWidgets() if isinstance(w, QMainWindow)]
        if len(main_windows) == 1 and main_windows[0] == self:
            self._confirm_quit(callback=do_close)
        else:
            do_close()

    @Slot(bool)
    def quit_main_window(self, checked):
        self._confirm_quit(callback=self.app.quit)

    def _confirm_quit(self, callback):
        if not config.CONFIRM_QUIT:
            callback()
        else:
            quit_message = QMessageBox.question(
                self, 'Quitting Application', 'Exit Application?',
                QMessageBox.Yes | QMessageBox.No)
            if quit_message == QMessageBox.Yes:
                callback()

    def show_macro_window(self):
        macro_window = MacroWindow(self)
        macro_window.show()

    def add_menu_items(self):
        # create the custom menu with user given items
        if not isinstance(self.display_widget(), Display):
            return

        # Only provide the view help menu option if an associated help file has been loaded
        if self.display_widget().help_window is None:
            self.ui.actionShow_Help.setVisible(False)

        items = self.display_widget().menu_items()
        if len(items) == 0:
            self.ui.menuCustomActions.menuAction().setVisible(False)
        else:
            self.create_menu(self.ui.menuCustomActions, items)

        # connect custom save, save as, and load functions
        file_menu_items = self.display_widget().file_menu_items()
        if len(file_menu_items) != 0:
            valid_keys = ('save', 'save_as', 'load')
            ui_actions = (self.ui.actionSave, self.ui.actionSave_As, self.ui.actionLoad)
            action_dict = dict(zip(valid_keys, ui_actions))
            # iterate through user given keys, which need to match the possible keys
            for key in file_menu_items.keys():
                if key.lower() in valid_keys:
                    ui_action = action_dict[key.lower()]
                    fileAction = file_menu_items[key]
                    # track shortcut status for callability verification below
                    has_shortcut = isinstance(fileAction, tuple)
                    if has_shortcut:
                        shortcut = fileAction[1]
                        fileAction = fileAction[0]
                        # track whether the action is callable after if statement
                        shortcut_action_callable = callable(fileAction)
                        if shortcut_action_callable:
                            PyDMMainWindow().add_shortcut(ui_action, shortcut)
                    if (has_shortcut and shortcut_action_callable) or callable(fileAction):
                        ui_action.setVisible(True)
                        ui_action.triggered.connect(fileAction)
                    else:
                        logger.error("Cannot add non callable object to menu.")
                else:
                    logger.error(f"File menu key not recognized, {key} action not added to file menu.")

    @staticmethod
    def create_menu(menu, items):
        # iterate through the items list and add to the menu
        for key in items.keys():
            val = items[key]
            # if there is a dictionary nested in the primary dictionary, create a new submenu, and call create_menu with that submenu as the parent menu
            if isinstance(val, dict):
                new_menu = menu.addMenu(key)
                PyDMMainWindow().create_menu(new_menu, val)
            # otherwise, add the key and it's relevant function to the menu
            else:
                action = menu.addAction(key)
                if isinstance(val, tuple):
                    shortcut = val[1]
                    val = val[0]
                    PyDMMainWindow().add_shortcut(action, shortcut)
                if callable(val):
                    action.triggered.connect(val)
                else:
                    menu.removeAction(key)
                    logger.error("Cannot add non callable object to menu.")

    @staticmethod
    def add_shortcut(action, shortcut):
        # set the shortcut to the given action
        action.setShortcutContext(Qt.ApplicationShortcut)
        action.setShortcut(QKeySequence(shortcut))



# Unit Tests for the PyDMSpinBox widget class


import pytest

from qtpy.QtWidgets import QApplication
from qtpy.QtGui import QKeyEvent
from qtpy.QtCore import QEvent, Qt

from ...widgets.spinbox import PyDMSpinbox
from ...tests.widgets.test_lineedit import find_action_from_menu


# --------------------
# POSITIVE TEST CASES
# --------------------

def test_construct(qtbot):
    """
    Test the construction of the widget.

    Expectations:
    All the default values are correctly assigned.

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    """
    pydm_spinbox = PyDMSpinbox()
    qtbot.addWidget(pydm_spinbox)

    assert pydm_spinbox.valueBeingSet is False
    assert pydm_spinbox.isEnabled() is False
    assert pydm_spinbox._show_step_exponent is True
    assert pydm_spinbox.step_exponent == 0
    assert pydm_spinbox.decimals() == 0
    assert pydm_spinbox.app == QApplication.instance()
    assert pydm_spinbox.isAccelerated() is True
    assert pydm_spinbox._write_on_press is False


@pytest.mark.parametrize("first_key_pressed, second_key_pressed, keys_pressed_expected_results", [
    (Qt.Key_Left, Qt.Key_Right, (2, 1)),
    (Qt.Key_Right, Qt.Key_Left, (-2, -1)),
])
def test_key_press_event(qtbot, signals, monkeypatch, first_key_pressed, second_key_pressed,
                         keys_pressed_expected_results):
    """
    Test the widget's handling of the key press events.

    Expectations:
    The step exponent value of the spin box will change as the Left Arrow or Right Arrow is pressed.

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    signals : fixture
        The signals fixture, which provides access signals to be bound to the appropriate slots
    monkeypatch : fixture
        To override default behaviors
    first_key_pressed : Qt.Key
        The first key to press to change the spinbox's step exponent value (increase or decrease)
    second_key_pressed: Qt.Key
        The second key to press to change the spinbox's step exponent value (increase or decrease)
    keys_pressed_expected_results : tuple
        The new value after the first key, and then that after the second key, is pressed
    """
    pydm_spinbox = PyDMSpinbox()
    qtbot.addWidget(pydm_spinbox)
    pydm_spinbox.setEnabled(True)

    with qtbot.waitExposed(pydm_spinbox):
        pydm_spinbox.show()

    pydm_spinbox.step_exponent = 0
    pydm_spinbox.precisionFromPV = True
    signals.prec_signal[int].connect(pydm_spinbox.precisionChanged)
    signals.prec_signal[int].emit(3)

    INIT_SPINBOX_VALUE = 1.2
    signals.new_value_signal[float].connect(pydm_spinbox.channelValueChanged)
    signals.new_value_signal[float].emit(INIT_SPINBOX_VALUE)

    pydm_spinbox.setFocus()

    def wait_focus():
        return pydm_spinbox.hasFocus()
    qtbot.waitUntil(wait_focus, timeout=5000)

    def press_key_and_verify(key_pressed, key_mod, key_press_count, expected_exp, expected_value):
        if key_mod != Qt.NoModifier:
            # Monkeypatch the Control flag because in this test, we don't properly have the app QApplication instance
            monkeypatch.setattr(QApplication, "queryKeyboardModifiers", lambda *args: key_mod)
        else:
            monkeypatch.setattr(QApplication, "queryKeyboardModifiers", lambda *args: Qt.NoModifier)

        for i in range(0, key_press_count):
            pydm_spinbox.keyPressEvent(QKeyEvent(QEvent.KeyPress, key_pressed, key_mod))
        assert pydm_spinbox.step_exponent == expected_exp

        signals.send_value_signal[float].connect(signals.receiveValue)
        signals.send_value_signal[float].connect(pydm_spinbox.channelValueChanged)

        # Send out the key event to update the value based on the step exponent change
        pydm_spinbox.keyPressEvent(QKeyEvent(QEvent.KeyPress, Qt.Key_Return, Qt.NoModifier))
        assert pydm_spinbox.value == expected_value

    # Send out the first key event to effect the change of the step exponent
    press_key_and_verify(first_key_pressed, Qt.ControlModifier, 2, keys_pressed_expected_results[0],
                         INIT_SPINBOX_VALUE)

    # Send out the second key press event once to check if it can change the step exponent
    press_key_and_verify(second_key_pressed, Qt.ControlModifier, 1, keys_pressed_expected_results[1],
                         INIT_SPINBOX_VALUE)

    # Make sure the widget can process the UpArrow, DownArrow, PageUp, and PageDown keys
    pydm_spinbox.keyPressEvent(QKeyEvent(QEvent.KeyPress, Qt.Key_Up, Qt.ControlModifier))
    pydm_spinbox.keyPressEvent(QKeyEvent(QEvent.KeyPress, Qt.Key_Down, Qt.ControlModifier))
    pydm_spinbox.keyPressEvent(QKeyEvent(QEvent.KeyPress, Qt.Key_PageUp, Qt.ControlModifier))
    pydm_spinbox.keyPressEvent(QKeyEvent(QEvent.KeyPress, Qt.Key_PageDown, Qt.ControlModifier))


def test_widget_ctx_menu(qtbot):
    """
    Test to make sure the widget's context menu contains the specific action for the widget.

    Expectations:
    All specific actions for the widget's context menu are present in that menu.

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    """
    pydm_spinbox = PyDMSpinbox()
    qtbot.addWidget(pydm_spinbox)

    menu = pydm_spinbox.widget_ctx_menu()
    action_menu = menu.menuAction().menu()
    assert find_action_from_menu(action_menu, "Toggle Show Step Size")


@pytest.mark.parametrize("step_exp", [
    10,
    1,
    0.01,
    0,
    -10,
    -0.001
])
def test_update_step_size(qtbot, step_exp):
    """
    Test the incrementing of the widget's step exponent.

    Expectations:
    The widget's step exponent is incremented correctly, by verifying the value returned by the Qt API singleStep() is
    ten times the step exponent value.

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    step_exp : int
        The new step exponent to set
    """
    pydm_spinbox = PyDMSpinbox()
    qtbot.addWidget(pydm_spinbox)

    pydm_spinbox.step_exponent = step_exp
    pydm_spinbox.update_step_size()

    assert pydm_spinbox.singleStep() == 10 ** step_exp


@pytest.mark.parametrize("show_unit, new_unit, step_exp, show_step_exp", [
    (True, "mJ", 0.01, True),
    (True, "light years", 10, True),
    (True, "s", -0.001, True),
    (True, "ms", 1, False),
    (False, "ns", 0.01, True),
    (False, "light years", 1, False),
])
def test_update_format_string(qtbot, signals, show_unit, new_unit, step_exp, show_step_exp):
    """
    Test the widget's capability of updating the format string when various unit and step exponent paramaters are
    updated. This method also test showStepExponent property and setter.

    Expectations:
    The format string is correctly produced.

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    signals : fixture
        The signals fixture, which provides access signals to be bound to the appropriate slots
    show_unit : bool
        True if the unit is to be displayed on the widget; False if otherwise
    new_unit : str
        The new unit to be displayed on the widget
    step_exp : int
        The step exponent for widget's values
    show_step_exp : bool
        True if the step exponent is to be displayed on the widget; False if otherwise
    """
    pydm_spinbox = PyDMSpinbox()
    qtbot.addWidget(pydm_spinbox)

    pydm_spinbox.step_exponent = step_exp
    pydm_spinbox.showStepExponent = show_step_exp
    assert pydm_spinbox.showStepExponent == show_step_exp

    INIT_SPINBOX_VALUE = 1.2
    signals.new_value_signal[float].connect(pydm_spinbox.channelValueChanged)
    signals.new_value_signal[float].emit(INIT_SPINBOX_VALUE)

    units = ""
    pydm_spinbox.showUnits = show_unit
    if show_unit:
        signals.unit_signal[str].connect(pydm_spinbox.unitChanged)
        signals.unit_signal.emit(new_unit)
        units = " {}".format(pydm_spinbox._unit)

        assert pydm_spinbox._unit == new_unit
    else:
        assert pydm_spinbox._unit == ""

    if show_step_exp:
        current_suffix = pydm_spinbox.suffix()
        expected_suffix = "{0} Step: 1E{1}".format(units, pydm_spinbox.step_exponent)
        assert current_suffix == expected_suffix
        assert pydm_spinbox.toolTip() == ""
    else:
        current_suffix = pydm_spinbox.suffix()
        expected_tooltip = "Step: 1E{0:+d}".format(pydm_spinbox.step_exponent)

        assert current_suffix == units
        assert pydm_spinbox.lineEdit().toolTip() == expected_tooltip


@pytest.mark.parametrize("init_value, user_typed_value, precision", [
    (123, 456, 3),
    (1.23, 4.56, 2),
    (1.23, 5, 0),
    (0, 12.3, 2),
    (-1.23, 4.6, 1)
])
def test_send_value(qtbot, signals, init_value, user_typed_value, precision):
    """
    Test sending the value from the widget to the channel. This method tests value_changed() and precision_changed().

    Expectations:
    The correct value is sent to the channelValueChanged slot.

    Parameters
    ----------
    qtbot : fixture
        pytest-qt window for widget test
    signals : fixture
        The signals fixture, which provides access signals to be bound to the appropriate slots
    init_value : float
        The initial value assigned to the spinbox
    user_typed_value : float
        The new value sent to the spinbox via the channel
    precision : int
        The decimal decision for the spinbox's value
    """
    pydm_spinbox = PyDMSpinbox()
    qtbot.addWidget(pydm_spinbox)

    pydm_spinbox.setValue(init_value)

    pydm_spinbox.precisionFromPV = True
    signals.prec_signal[type(precision)].connect(pydm_spinbox.precisionChanged)
    signals.prec_signal[type(precision)].emit(precision)

    signal_type = type(user_typed_value)
    signals.new_value_signal[signal_type].connect(pydm_spinbox.channelValueChanged)
    signals.new_value_signal[signal_type].emit(user_typed_value)

    # Besides receiving the new channel value, simulate the update of the new value to the widget by connecting the
    # channelValueChanged slot to the same signal
    signals.send_value_signal[signal_type].connect(signals.receiveValue)
    signals.send_value_signal[signal_type].connect(pydm_spinbox.channelValueChanged)
    pydm_spinbox.send_value()

    assert pydm_spinbox.value == user_typed_value


@pytest.mark.parametrize("which_limit, new_limit, user_defined_limits", [
    ("UPPER", 123.456, False),
    ("LOWER", 12.345, False),
    ("UPPER", 987.654, True),
    ("LOWER", 9.321, True)
])
def test_ctrl_limit_changed(qtbot, signals, which_limit, new_limit, user_defined_limits):
    """
    Test the upper and lower limit settings.

    Expectations:
        The upper or lower limit can be emitted and subsequently read correctly.

    Parameters
    ----------
    qtbot : fixture
        pytest-qt window for widget test
    signals : fixture
        The signals fixture, which provides access signals to be bound to the appropriate slots
    which_limit : str
        "UPPER" if the new value is intended for the upper limit, "LOWER" for the lower limit
    new_limit : float
        The new limit value
    user_defined_limits : bool
        True if the spinbox should set its range based on user defined values, False if it should take its
        range from the PV itself
    """
    pydm_spinbox = PyDMSpinbox()
    qtbot.addWidget(pydm_spinbox)
    pydm_spinbox.userDefinedLimits = user_defined_limits
    pydm_spinbox.userMinimum = -10.5
    pydm_spinbox.userMaximum = 10.5
    pydm_spinbox.precisionFromPV = False
    pydm_spinbox.precision = 3

    if which_limit == "UPPER":
        signals.upper_ctrl_limit_signal[type(new_limit)].connect(pydm_spinbox.upperCtrlLimitChanged)
        signals.upper_ctrl_limit_signal[type(new_limit)].emit(new_limit)

        assert pydm_spinbox.get_ctrl_limits()[1] == new_limit
        if not user_defined_limits:
            # Not user_defined_limits means the range of the spinbox should stay in sync with the PV
            assert pydm_spinbox.maximum() == new_limit
        else:
            # Otherwise while we still store the limits on the base widget, the spinbox remains at the user-set values
            assert pydm_spinbox.maximum() == 10.5
    elif which_limit == "LOWER":
        signals.lower_ctrl_limit_signal[type(new_limit)].connect(pydm_spinbox.lowerCtrlLimitChanged)
        signals.lower_ctrl_limit_signal[type(new_limit)].emit(new_limit)

        assert pydm_spinbox.get_ctrl_limits()[0] == new_limit
        if not user_defined_limits:
            assert pydm_spinbox.minimum() == new_limit
        else:
            assert pydm_spinbox.minimum() == -10.5


def test_reset_limits(qtbot):
    """
    Test that when reset_limits() is called, the range of the spinbox is set as expected
    """
    # Set up a spinbox with some initial user set limits
    pydm_spinbox = PyDMSpinbox()
    qtbot.addWidget(pydm_spinbox)
    pydm_spinbox.userDefinedLimits = True
    pydm_spinbox.userMinimum = -10.5
    pydm_spinbox.userMaximum = 10.5
    pydm_spinbox.precisionFromPV = False
    pydm_spinbox.precision = 1

    # Also set a couple limits based on info we could have gotten back from the channel
    pydm_spinbox._lower_ctrl_limit = -5
    pydm_spinbox._upper_ctrl_limit = 5

    # This first call to reset_limits should essentially do nothing as nothing has happened
    # that would cause the range of the spinbox to change
    pydm_spinbox.reset_limits()
    assert pydm_spinbox.minimum() == -10.5
    assert pydm_spinbox.maximum() == 10.5

    # Turning off user defined limits should now cause the range to be set by the values received from the channel
    pydm_spinbox.userDefinedLimits = False
    assert pydm_spinbox.minimum() == -5
    assert pydm_spinbox.maximum() == 5

    # Finally flip it back to ensure it switches back to the user-defined range as expected
    pydm_spinbox.userDefinedLimits = True
    assert pydm_spinbox.minimum() == -10.5
    assert pydm_spinbox.maximum() == 10.5


@pytest.mark.parametrize("key_pressed, initial_spinbox_value, expected_result, write_on_press", [
    (Qt.Key_Up, 1, 2, True),
    (Qt.Key_Down, 1, 0, True),
    (Qt.Key_Up, 1, 1, False),
    (Qt.Key_Down, 1, 1, False),
])
def test_write_on_press(qtbot, signals, monkeypatch, key_pressed, initial_spinbox_value,
                        expected_result, write_on_press):
    """
    Test sending the value from the widget to the channel on key press when writeOnPress enabled.

    Expectations:
    The correct value is sent to the channelValueChanged slot on key press and no value is send when disabled.

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    signals : fixture
        The signals fixture, which provides access signals to be bound to the appropriate slots
    monkeypatch : fixture
        To override default behaviors
    key_pressed : Qt.Key
        The key to press to change the spinbox's step exponent value (increase or decrease)
    inital_spinbox_value: float
        Initial value assigned to spinbox
    expected_result : float
        Value expected after key press
    write_on_press: bool
        Whether or not to enable write on press
    """
    pydm_spinbox = PyDMSpinbox()
    qtbot.addWidget(pydm_spinbox)

    with qtbot.waitExposed(pydm_spinbox):
        pydm_spinbox.show()

    pydm_spinbox.step_exponent = 0
    pydm_spinbox.precisionFromPV = True
    signals.prec_signal[int].connect(pydm_spinbox.precisionChanged)
    signals.prec_signal[int].emit(3)

    # set up signals
    pydm_spinbox.send_value_signal[float].connect(signals.receiveValue)
    signals.new_value_signal[float].connect(pydm_spinbox.channelValueChanged)
    signals.new_value_signal[float].emit(initial_spinbox_value)

    def press_key_and_verify(key_pressed, key_mod, expected_value, write_on_press):
        # reset value to initial value
        pydm_spinbox.value_changed(initial_spinbox_value)
        pydm_spinbox.send_value()

        #  use modifier if left/right keys
        if key_mod != Qt.NoModifier:
            # Monkeypatch the Control flag because in this test, we don't properly have the app QApplication instance
            monkeypatch.setattr(QApplication, "queryKeyboardModifiers", lambda *args: key_mod)
        else:
            monkeypatch.setattr(QApplication, "queryKeyboardModifiers", lambda *args: Qt.NoModifier)

        pydm_spinbox.writeOnPress = write_on_press
        pydm_spinbox.keyPressEvent(QKeyEvent(QEvent.KeyPress, key_pressed, key_mod))

        # Check for change if write on press enabled
        if pydm_spinbox.writeOnPress:
            assert signals.value == expected_value

        # Check for no change if disabled
        else:
            assert signals.value == initial_spinbox_value

    # check the key press with write_on_press
    press_key_and_verify(key_pressed, Qt.NoModifier, expected_result, write_on_press)

    # check that changing the exponent does not modify
    press_key_and_verify(Qt.Key_Left, Qt.ControlModifier, initial_spinbox_value, write_on_press)
    press_key_and_verify(Qt.Key_Right, Qt.ControlModifier, initial_spinbox_value, write_on_press)



import collections
import functools
import inspect
import logging
import os
from typing import Any, Dict, Generator, List, Optional, Union

import entrypoints
from qtpy.QtWidgets import QMenu, QWidget

from ..config import ENTRYPOINT_EXTERNAL_TOOL, EXTERNAL_TOOL_SUFFIX
from ..utilities import import_module_by_filename, log_failures
from .tools import ExternalTool

logger = logging.getLogger(__name__)
ext_tools: Dict[str, Any] = {}
_ext_tools_loaded: bool = False


def _is_valid_external_tool_class(obj: Any) -> bool:
    """Is the object a valid external tool?"""
    return (
        inspect.isclass(obj)
        and issubclass(obj, ExternalTool)
        and obj is not ExternalTool
    )


@log_failures(logger, explanation="Failed to load External Tool: {args[0]}")
def _get_tools_from_source(source_filename: str) -> List[ExternalTool]:
    """
    For a given source filename, find ExternalTool classes.

    Parameters
    ----------
    source_filename : str
        The source code filename.

    Returns
    -------
    tools : list of ExternalTool
        Instantiated external tools.

    Raises
    ------
    ValueError
        If no subclassed external tools are found.
    """
    module = import_module_by_filename(source_filename)
    classes = list(
        set(
            obj
            for _, obj in inspect.getmembers(module)
            if _is_valid_external_tool_class(obj)
        )
    )

    if not classes:
        raise ValueError(
            f"Invalid File Format. {source_filename} has no class inheriting "
            "from ExternalTool. Nothing to open at this time."
        )
    return [c() for c in classes]


@log_failures(logger, explanation="Failed to load External Tool: {args[0]}")
def install_external_tool(tool: Union[str, ExternalTool]) -> None:
    """
    Install an External Tool at the PyDMApplication and add it to the main
    window Tools menu.

    Parameters
    ----------
    tool : str or pydm.tools.ExternalTool
        The full path to a file containing a ExternalTool definition
        or an Instance of an ExternalTool.
    """
    global ext_tools

    if isinstance(tool, str):
        objects = _get_tools_from_source(tool) or []
    else:
        objects = [tool]

    for tool_obj in objects:
        if not isinstance(tool_obj, ExternalTool):
            raise ValueError(
                f"Invalid tool found: {tool}. String or ExternalTool "
                f"expected."
            )

        if tool_obj.group is not None and tool_obj.group:
            if tool_obj.group not in ext_tools:
                ext_tools[tool_obj.group] = {}
            ext_tools[tool_obj.group][tool_obj.name] = tool_obj
        else:
            ext_tools[tool_obj.name] = tool_obj

    ext_tools = collections.OrderedDict(sorted(ext_tools.items()))
    for ext_tool_name in ext_tools:
        if isinstance(ext_tools[ext_tool_name], dict):
            ext_tools[ext_tool_name] = collections.OrderedDict(
                sorted(ext_tools[ext_tool_name].items())
            )


def assemble_tools_menu(
    parent_menu: QMenu,
    clear_menu: bool = False,
    widget_only: bool = False,
    widget: Optional[QWidget] = None,
    **kwargs
) -> None:
    """
    Assemble the Tools menu for a given parent menu.

    Parameters
    ----------
    parent_menu : QMenu
        The main menu item to hold the tools menu tree.
    clear_menu : bool, optional
        Whether of not we should clear the menu before adding the tools.
    widget_only : bool, optional
        Whether or not generate only the menu for widgets compatible
        tools. This should be True when creating the menu for the
        PyDMWidgets and False for most of the other cases.
    widget : QWidget, optional
        The widget for which the menu is being assembled. This allow for the
        tools to filter if they are compatible or not with the widget based
        on properties, types and etc. Default is None which means all tools
        will be assembled.
    **kwargs :
        Parameters sent directly to the `call` method of the ExternalTool
        instance. In general this dict is composed by `channels` which
        is a list and `sender` which is a QWidget.
    """
    def assemble_action(menu, tool_obj):
        if tool_obj.icon is not None:
            action = menu.addAction(tool_obj.name)
            action.setIcon(tool_obj.icon)
        else:
            action = menu.addAction(tool_obj.name)
        action.triggered.connect(functools.partial(tool_obj.call, **kwargs))

    load_external_tools()

    if clear_menu:
        parent_menu.clear()
    else:
        parent_menu.addSeparator()

    for k, v in ext_tools.items():
        if isinstance(v, dict):
            m = QMenu(k, parent=parent_menu)
            should_create_menu = False
            for _, t in v.items():
                if widget_only:
                    if not t.use_with_widgets:
                        continue
                    if widget is not None and not t.is_compatible_with(widget):
                        logger.debug(
                            "Skipping tool %s as it is incompatible with "
                            "widget %s.",
                            t.name,
                            widget,
                        )
                        continue
                elif not t.use_without_widget:
                    continue
                assemble_action(m, t)
                should_create_menu = True
            if should_create_menu:
                parent_menu.addMenu(m)
        else:
            if widget_only and not v.use_with_widgets:
                continue
            assemble_action(parent_menu, v)


def get_entrypoint_tools() -> Generator[ExternalTool, None, None]:
    """
    Yield all external tool classes specified by entrypoints.

    Uses ``entrypoints`` to find packaged external tools in packages that
    configure the ``pydm.tool`` entrypoint.
    """
    for entry in entrypoints.get_group_all(ENTRYPOINT_EXTERNAL_TOOL):
        logger.debug("Found external tool entrypoint: %s", entry.name)
        try:
            tool_cls = entry.load()
        except Exception as ex:
            logger.exception(
                "Failed to load %s entry %s: %s",
                ENTRYPOINT_EXTERNAL_TOOL, entry.name, ex
            )
            continue

        if not _is_valid_external_tool_class(tool_cls):
            logger.warning(
                "Invalid external tool class specified in entrypoint "
                "%s: %s",
                entry.name, tool_cls
            )
            continue

        yield tool_cls()


def get_tools_from_path() -> Generator[ExternalTool, None, None]:
    """Yield all external tool classes specified by PYDM_TOOLS_PATH."""
    tools_path = os.getenv("PYDM_TOOLS_PATH", None)

    if not tools_path:
        logger.debug(
            "External Tools not loaded from PYDM_TOOLS_PATH as no path "
            "was specified."
        )
        return

    logger.debug("Looking for external tools at: %s", tools_path)
    for loc in tools_path.split(os.pathsep):
        for root, _, files in os.walk(loc):
            for name in files:
                if name.endswith(EXTERNAL_TOOL_SUFFIX):
                    tool_path = os.path.join(root, name)
                    logger.debug("Found tool in %s", tool_path)
                    yield from _get_tools_from_source(tool_path)


def load_external_tools():
    """
    Loads all the external tools available.

    1. Uses ``PYDM_TOOLS_PATH`` environment variable. Searches all directories
       and subdirectories for Python source code that matches the pattern
       ``*_tool.py`` which contain classes that inherit from
       :class:`pydm.tools.ExternalTool`.
    2. Uses ``entrypoints`` to find packaged external tools in packages that
       configure the ``pydm.tool`` entrypoint.

    If called previously, this function is a no-operation.
    """
    global _ext_tools_loaded

    if _ext_tools_loaded:
        return

    _ext_tools_loaded = True

    for tool in get_tools_from_path():
        install_external_tool(tool)

    for tool in get_entrypoint_tools():
        install_external_tool(tool)



import io
import re
import six
from string import Template
import json

# Macro parsing states
PRE_NAME = 0
IN_NAME = 1
PRE_VAL = 2
IN_VAL = 3


def substitute_in_file(file_path, macros):
    """
    Substitute the macros given by ${name} at the given file with the entries on the `macros` dictionary.

    Parameters
    ----------
    file_path : str
        The path to the file in which to substitute
    macros : dict
        Dictionary containing macro name as key and value as what will be substituted.
    Returns
    -------
    file : io.StringIO
        File-like object with the proper substitutions.
    """
    template = template_for_file(file_path)
    return replace_macros_in_template(template, macros)


def replace_macros_in_template(template, macros):
    curr_template = template
    prev_template = Template("")
    expanded_text = ""
    # Escape any single or double quotes to ensure macro substitution results in valid python code
    # when replaced (e.g. xterm -e 'echo hi')
    macros = {key: re.sub(r'(?<!\\)"', '\\"', re.sub(r"(?<!\\)'", "\\'", value)) if isinstance(value, str) else value
              for key, value in macros.items()}

    for i in range(100):
        expanded_text = curr_template.safe_substitute(macros)
        if curr_template.template == prev_template.template:
            break
        prev_template = curr_template
        curr_template = Template(expanded_text)
    return io.StringIO(six.text_type(expanded_text))


def template_for_file(file_path):
    with open(file_path) as orig_file:
        text = Template(orig_file.read())
    return text


def parse_macro_string(macro_string):
    """Parses a macro string and returns a dictionary.
    First, this method attempts to parse the string as JSON.
    If that fails, it attempts to parse it as an EPICS-style
    macro string.  The parsing algorithm for that case is very
    closely based on macParseDefns in libCom/macUtil.c"""
    if not macro_string:
        return {}

    macro_string = str(macro_string)

    try:
        macros = json.loads(macro_string)
        return macros
    except ValueError:
        if "=" not in macro_string:
            raise ValueError("Could not parse macro argument as JSON.")
        macros = {}
        state = PRE_NAME
        quote = None
        name_start = None
        name_end = None
        val_start = None
        val_end = None
        for (i,c) in enumerate(macro_string):
            if quote:
                if c == quote:
                    quote = False
            elif c == "'" or c == '"':
                quote = c
                continue
            escape = macro_string[i-1] == "\\"
            if state == PRE_NAME:
                if (not quote) and (not escape) and (c.isspace() or c == ","):
                    continue
                name_start = i
                state = IN_NAME
            elif state == IN_NAME:
                if quote or escape:
                    continue
                if c == "=" or c == ",":
                    name_end = i
                    state = PRE_VAL
            elif state == PRE_VAL:
                if (not quote) and (not escape) and c.isspace():
                    continue
                val_start = i
                state = IN_VAL
                if i == len(macro_string)-1:
                    val_end = i+1
            elif state == IN_VAL:
                if quote or escape:
                    continue
                if c == ",":
                    val_end = i
                    state = PRE_NAME
                elif i == len(macro_string)-1:
                    val_end = i+1
                    state = PRE_NAME
                else:
                    continue
            if not (None in (name_start, name_end, val_start, val_end)):
                key = macro_string[name_start:name_end].strip().replace("\\", "")
                val = macro_string[val_start:val_end].strip('"\'').replace("\\", "")
                macros[key] = val
                name_start = None
                name_end = None
                val_start = None
                val_end = None
                state = PRE_NAME
        return macros
    


from .base import PyDMWidget
from qtpy.QtGui import QColor, QPolygon, QPainter, QFontMetrics
from qtpy.QtWidgets import QFrame, QSizePolicy
from qtpy.QtCore import Qt, QPoint, Property, QSize
from qtpy.QtWidgets import QWIDGETSIZE_MAX
from .scale import QScale, PyDMScaleIndicator


class QScaleAlarmed(QScale):
    """
    Adds alarm regions and features for QScale.
    Additional configurable features include indicator type (bar/pointer),
    scale tick marks and orientation (horizontal/vertical).
    Parameters
    ----------
    parent : QWidget
        The parent widget for the PyDMAnalogIndicator
    """

    def __init__(self, parent=None):
        super(QScaleAlarmed, self).__init__(parent)
        self._lower_minor_alarm = 0
        self._upper_minor_alarm = 0
        self._lower_major_alarm = 0
        self._upper_major_alarm = 0

        """
        self._cannot_draw_*_flag variables are used so draw errors only printout once.
        False = last time the paint event was called the element was able to draw
        True = last time the paint event was called the element was not able to draw because of an error
        """
        self._cannot_draw_background_flag = False
        self._cannot_draw_major_alarm_region_flag = False
        self._cannot_draw_minor_alarm_region_flag = False
        self._cannot_draw_ticks_flag = False
        self._cannot_draw_indicator_flag = False

        self._bg_color = QColor('lightblue')
        self._minor_alarm_region_color = QColor('white')
        self._minor_alarm_color = QColor('yellow')
        self._major_alarm_region_color = QColor('grey')
        self._major_alarm_color = QColor('red')
        self._bg_size_rate = 0.5  # from 0 to 1
        self._scale_height = 40

        self._show_ticks = False

        self.set_position()

    def adjust_transformation(self):
        """
        This method sets parameters for the widget transformations
        (needed to for orientation, flipping and appearance inversion).
        Rewritten to expand scale when in horizontal position and value
        displayed on left/right
        also helpful for use in layouts
        """
        self.setMaximumSize(QWIDGETSIZE_MAX,
                            QWIDGETSIZE_MAX)  # Unset fixed size
        if self._orientation == Qt.Horizontal:
            self._widget_width = self.width()
            self._widget_height = self.height()
            self._painter_translation_y = 0
            self._painter_rotation = 0
            # expands scale in horizontal position
            self.setSizePolicy(QSizePolicy.Expanding,
                               QSizePolicy.Fixed)
            self.setFixedHeight(self._scale_height)
        elif self._orientation == Qt.Vertical:
            # Invert dimensions for paintEvent()
            self._widget_width = self.height()
            self._widget_height = self.width()
            self._painter_translation_y = self._widget_width
            self._painter_rotation = -90
            self.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Expanding)
            self.setFixedWidth(self._scale_height)

        if self._inverted_appearance:
            self._painter_translation_x = self._widget_width
            self._painter_scale_x = -1
        else:
            self._painter_translation_x = 0
            self._painter_scale_x = 1

        if self._flip_scale:
            self._flip_traslation_y = self._widget_height
            self._flip_scale_y = -1
        else:
            self._flip_traslation_y = 0
            self._flip_scale_y = 1

    def draw_ticks(self):
        """
        Draw tick marks on the scale.
        Rewrote to draw them in a different place.
        """
        if not self._show_ticks:
            return
        self.set_tick_pen()
        self._painter.setPen(self._tick_pen)
        division_size = self._widget_width / self._num_divisions
        tick_y0 = self._scale_height * self._bg_size_rate
        tick_yf = tick_y0 + self._scale_height * self._tick_size_rate * .25
        for i in range(self._num_divisions + 1):
            x = i * division_size
            self._painter.drawLine(x, tick_y0, x, tick_yf)  # x1, y1, x2, y2

    """
    Currently unused, needs pulling in a second PV to be fully implemented
    def draw_set_point(self):

        #Draw a pointer as indicator of current value.

        self.set_position()
        if self.position < 0 or self.position > self._widget_width:
            return
        self._painter.setPen(Qt.transparent)
        self._painter.setBrush(self._indicator_color)
        pointer_width = self._pointer_width_rate * self._widget_width
        pointer_height = self._bg_size_rate * self._widget_height
        points = [
            QPoint(self.position, 15),
            QPoint(self.position + 0.5*pointer_width, 0.5*pointer_height+15),
            QPoint(self.position, pointer_height),
            QPoint(self.position - 0.5*pointer_width, 0.5*pointer_height+15)
        ]
        self._painter.drawPolygon(QPolygon(points))
    """

    def draw_indicator(self):
        """
        Draw the indicator for current value.
        """
        self.set_position()
        if self.position < 0 or self.position > self._widget_width:
            return
        self._painter.setPen(Qt.transparent)
        self._painter.setBrush(self._indicator_color)
        pointer_width = self._pointer_width_rate * self._widget_width
        pointer_height = self._bg_size_rate * self._scale_height * 1.5
        points = [
            QPoint(self.position + 0.5 * pointer_width, 0),
            QPoint(self.position - 0.5 * pointer_width, 0),
            QPoint(self.position, pointer_height)
        ]
        self._painter.drawPolygon(QPolygon(points))

    def draw_background(self):
        """
        Draw the background of the scale.
        """
        self._painter.setPen(Qt.black)
        self._painter.setBrush(self._bg_color)
        pointer_height = self._bg_size_rate * self._scale_height
        bg_width = self._widget_width
        bg_height = self._bg_size_rate * self._widget_height - 2
        self._painter.drawRect(0, pointer_height, bg_width, bg_height)

    def draw_minor_alarm_region(self):
        """
        Draw the minor alarm areas on the scale
        """

        self._painter.setPen(Qt.black)
        self._painter.setBrush(self._minor_alarm_region_color)
        lower_minor_alarm_width = self.calculate_position_for_value(self._lower_minor_alarm)

        upper_minor_alarm_start = self.calculate_position_for_value(self._upper_minor_alarm)
        upper_minor_alarm_width = self._widget_width - upper_minor_alarm_start
        pointer_height = self._bg_size_rate * self._scale_height
        minor_alarm_height = self._bg_size_rate * self._widget_height - 2

        """
        sets the pen color to alarm if the value is in the lower minor alarm region
        """
        if self._value <= self._lower_minor_alarm and self._value > self._lower_major_alarm:
            self._painter.setBrush(self._minor_alarm_color)
        else:
            self._painter.setBrush(self._minor_alarm_region_color)
        if self._lower_minor_alarm > self._lower_limit:
            self._painter.drawRect(0, pointer_height, lower_minor_alarm_width, minor_alarm_height)
        """
        sets the pen color to alarm if the value is in the upper minor alarm region
        """
        if self._value >= self._upper_minor_alarm and self._value < self._upper_major_alarm:
            self._painter.setBrush(self._minor_alarm_color)
        else:
            self._painter.setBrush(self._minor_alarm_region_color)
        if self._upper_minor_alarm < self._upper_limit:
            self._painter.drawRect(upper_minor_alarm_start, pointer_height, upper_minor_alarm_width, minor_alarm_height)

    def draw_major_alarm_region(self):
        """
        Draw the minor alarm areas on the scale
        """

        self._painter.setPen(Qt.black)
        self._painter.setBrush(self._major_alarm_region_color)
        lower_major_alarm_width = self.calculate_position_for_value(self._lower_major_alarm)

        upper_major_alarm_start = self.calculate_position_for_value(self._upper_major_alarm)
        upper_major_alarm_width = self._widget_width - upper_major_alarm_start
        pointer_height = self._bg_size_rate * self._scale_height

        major_alarm_height = self._bg_size_rate * self._widget_height - 2

        """
        sets the pen color to alarm if the value is in the lower major alarm region
        """
        if self._value <= self._lower_major_alarm:
            self._painter.setBrush(self._major_alarm_color)
        else:
            self._painter.setBrush(self._major_alarm_region_color)
        # makes sure alarm value is in range
        if self._lower_major_alarm > self._lower_limit:
            self._painter.drawRect(0, pointer_height, lower_major_alarm_width, major_alarm_height)

        """
        sets the pen color to alarm if the value is in the upper major alarm region
        """
        if self._value >= self._upper_major_alarm:
            self._painter.setBrush(self._major_alarm_color)
        else:
            self._painter.setBrush(self._major_alarm_region_color)
        # makes sure alarm value is in range
        if self._upper_major_alarm < self._upper_limit:
            self._painter.drawRect(upper_major_alarm_start, pointer_height, upper_major_alarm_width, major_alarm_height)

    def paintEvent(self, event):
        """
        Paint events are sent to widgets that need to update themselves,
        for instance when part of a widget is exposed because a covering
        widget was moved.
        Parameters
        ----------
        event : QPaintEvent
        """
        self.adjust_transformation()
        self._painter.begin(self)
        self._painter.translate(0, self._painter_translation_y)  # Draw vertically if needed
        self._painter.rotate(self._painter_rotation)
        self._painter.translate(self._painter_translation_x, 0)  # Invert appearance if needed
        self._painter.scale(self._painter_scale_x, 1)

        self._painter.translate(0, self._flip_traslation_y)  # Invert scale if needed
        self._painter.scale(1, self._flip_scale_y)

        self._painter.setRenderHint(QPainter.Antialiasing)

        """
        bad metadata or user input can cause designer or pydm to crash when drawing the widget, hence the try except block
        self._cannot_draw_*_flag variables are so that the errors only print once, otherwise they would print on each redraw
        flags are reset when the element can draw without error
        """
        try:
            self.draw_background()
        except Exception:
            if not self._cannot_draw_background_flag:
                print("Error: PyDMAlalogIndicator can't draw background, check upper and lower limits")
            self._cannot_draw_background_flag = True
        else:
            self._cannot_draw_background_flag = False
        try:
            if not self._upper_minor_alarm == self._lower_minor_alarm == 0:
                self.draw_minor_alarm_region()
        except Exception:
            if not self._cannot_draw_minor_alarm_region_flag:
                print("Error: PyDMAlalogIndicator can't draw minor alarm region, check minor alarm values and limits")
                self._cannot_draw_minor_alarm_region_flag = True
        else:
            self._cannot_draw_minor_alarm_region_flag = False
        try:
            if not self._upper_major_alarm == self._lower_major_alarm == 0:
                self.draw_major_alarm_region()
        except Exception:
            if not self._cannot_draw_major_alarm_region_flag:
                print("Error: PyDMAlalogIndicator can't draw major alarm region, check major alarm values and limits")
                self._cannot_draw_major_alarm_region_flag = True
        else:
            self._cannot_draw_major_alarm_region_flag = False
        try:
            self.draw_ticks()
        except Exception:
            if not self._cannot_draw_ticks_flag:
                print("Error: PyDMAlalogIndicator can't draw ticks")
                self._cannot_draw_ticks_flag = True
        else:
            self._cannot_draw_ticks_flag = False
        try:
            self.draw_indicator()
        except Exception:
            if not self._cannot_draw_indicator_flag:
                print("Error: PyDMAlalogIndicator can't draw_indicator")
                self._cannot_draw_indicator_flag = True
        else:
            self._cannot_draw_indicator_flag = False

        self._painter.end()

    def set_upper_minor_alarm(self, new_minor_alarm):
        """
        Set the scale upper minor alarm.
        Parameters
        ----------
        new_minor_alarm : float
            The upper minor alarm of the scale.
        """
        self._upper_minor_alarm = new_minor_alarm

    def set_lower_minor_alarm(self, new_minor_alarm):
        """
        Set the scale lower minor alarm.
        Parameters
        ----------
        new_minor_alarm : float
            The lower minor alarm of the scale.
        """
        self._lower_minor_alarm = new_minor_alarm

    def set_upper_major_alarm(self, new_major_alarm):
        """
        Set the scale upper major alarm.
        Parameters
        ----------
        new_major_alarm : float
            The upper major alarm of the scale.
        """
        self._upper_major_alarm = new_major_alarm

    def set_lower_major_alarm(self, new_major_alarm):
        """
        Set the scale lower major alarm.
        Parameters
        ----------
        new_major_alarm : float
            The lower major alarm of the scale.
        """
        self._lower_major_alarm = new_major_alarm

    def get_minor_alarm_region_color(self):
        return self._minor_alarm_region_color

    def set_minor_alarm_region_color(self, color):
        self._minor_alarm_region_color = color
        self.repaint()

    def get_minor_alarm_color(self):
        return self._minor_alarm_color

    def set_minor_alarm_color(self, color):
        self._minor_alarm_color = color
        self.repaint()

    def get_major_alarm_region_color(self):
        return self._major_alarm_region_color

    def set_major_alarm_region_color(self, color):
        self._major_alarm_region_color = color
        self.repaint()

    def get_major_alarm_color(self):
        return self._major_alarm_color

    def set_major_alarm_color(self, color):
        self._major_alarm_color = color
        self.repaint()

    def minimumSizeHint(self):
        fm = QFontMetrics(self.font())
        return QSize(fm.height(), fm.height())

    # reject some inherited things here


class PyDMAnalogIndicator(PyDMScaleIndicator):
    """
    A bar-shaped indicator for scalar value with support for Channels and
    more from PyDM.
    Configurable features include indicator type (bar/pointer), scale tick
    marks and orientation (horizontal/vertical).
    Parameters
    ----------
    parent : QWidget
        The parent widget for the Scale
    init_channel : str, optional
        The channel to be used by the widget.
    """

    def __init__(self, parent=None, init_channel=None):
        QFrame.__init__(self, parent)
        PyDMWidget.__init__(self, init_channel=init_channel)
        PyDMScaleIndicator.__init__(self, parent)
        self._show_limits = False
        self.scale_indicator = QScaleAlarmed()

        self._value_position = Qt.RightEdge
        self._minor_alarm_from_channel = True
        self._major_alarm_from_channel = True
        self._user_lower_minor_alarm = 0
        self._user_upper_minor_alarm = 0
        self._user_lower_major_alarm = 0
        self._user_upper_major_alarm = 0

        self.value_label.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)
        self.setup_widgets_for_orientation(Qt.Horizontal, False, False, self._value_position)
        self.lower_label.hide()
        self.upper_label.hide()

    def sizeHint(self):
        """
        makes default size nice when dragging into designer nice
        """
        return QSize(250, 70)

    def lower_warning_limit_changed(self, new_minor_alarm):
        """
        Callback updates the lower minor alarm boundary
        """
        super(PyDMAnalogIndicator, self).lower_warning_limit_changed(new_minor_alarm)
        if self.minorAlarmFromChannel:
            self.scale_indicator.set_lower_minor_alarm(new_minor_alarm)
            self.update_labels()

    def upper_warning_limit_changed(self, new_minor_alarm):
        """
        Callback updates the upper minor alarm boundary
        """
        super(PyDMAnalogIndicator, self).upper_warning_limit_changed(new_minor_alarm)
        if self.minorAlarmFromChannel:
            self.scale_indicator.set_upper_minor_alarm(new_minor_alarm)
            self.update_labels()

    def lower_alarm_limit_changed(self, new_major_alarm):
        """
        Callback updates the lower major alarm boundary
        """
        super(PyDMAnalogIndicator, self).lower_alarm_limit_changed(new_major_alarm)
        if self.majorAlarmFromChannel:
            self.scale_indicator.set_lower_major_alarm(new_major_alarm)
            self.update_labels()

    def upper_alarm_limit_changed(self, new_major_alarm):
        """
        Callback updates the upper major alarm boundary
        """
        super(PyDMAnalogIndicator, self).upper_alarm_limit_changed(new_major_alarm)
        if self.majorAlarmFromChannel:
            self.scale_indicator.set_upper_major_alarm(new_major_alarm)
            self.update_labels()

    # Reject barIndicator setter and getter

    @Property(QColor)
    def minorAlarmRegionColor(self):
        """
        The color of the scale background.
        Returns
        -------
        QColor
        """
        return self.scale_indicator.get_minor_alarm_region_color()

    @minorAlarmRegionColor.setter
    def minorAlarmRegionColor(self, color):
        """
        The color of the scale background.
        Parameters
        -------
        color : QColor
        """
        self.scale_indicator.set_minor_alarm_region_color(color)

    @Property(QColor)
    def minorAlarmColor(self):
        """
        The color of the scale background.
        Returns
        -------
        QColor
        """
        return self.scale_indicator.get_minor_alarm_color()

    @minorAlarmColor.setter
    def minorAlarmColor(self, color):
        """
        The color of the scale background.
        Parameters
        -------
        color : QColor
        """
        self.scale_indicator.set_minor_alarm_color(color)

    @Property(QColor)
    def majorAlarmRegionColor(self):
        """
        The color of the scale background.
        Returns
        -------
        QColor
        """
        return self.scale_indicator.get_major_alarm_region_color()

    @majorAlarmRegionColor.setter
    def majorAlarmRegionColor(self, color):
        """
        The color of the scale background.
        Parameters
        -------
        color : QColor
        """
        self.scale_indicator.set_major_alarm_region_color(color)

    @Property(QColor)
    def majorAlarmColor(self):
        """
        The color of the scale background.
        Returns
        -------
        QColor
        """
        return self.scale_indicator.get_major_alarm_color()

    @majorAlarmColor.setter
    def majorAlarmColor(self, color):
        """
        The color of the scale background.
        Parameters
        -------
        color : QColor
        """
        self.scale_indicator.set_major_alarm_color(color)

    @Property(bool)
    def minorAlarmFromChannel(self):
        """
        Whether or not the scale indicator should use the minor alarm information
        from the channel.
        Returns
        -------
        bool
        """
        return self._minor_alarm_from_channel

    @minorAlarmFromChannel.setter
    def minorAlarmFromChannel(self, checked):
        """
        Whether or not the scale indicator should use the minor alarm information
        from the channel.
        Parameters
        ----------
        checked : bool
            True to use the minor alarm from the Channel, False to use the user-defined
            values.
        """
        if self._minor_alarm_from_channel != checked:
            self._minor_alarm_from_channel = checked
            if checked:
                if self.lower_warning_limit:
                    self.scale_indicator.set_lower_minor_alarm(self.lower_warning_limit)
                if self.upper_warning_limit:
                    self.scale_indicator.set_upper_minor_alarm(self.upper_warning_limit)
            else:
                self.scale_indicator.set_lower_minor_alarm(self._user_lower_minor_alarm)
                self.scale_indicator.set_upper_minor_alarm(self._user_upper_minor_alarm)
            self.update_labels()

    @Property(float)
    def userUpperMinorAlarm(self):
        """
        The user-defined upper minor alarm for the scale.
        Returns
        -------
        float
        """
        return self._user_upper_minor_alarm

    @userUpperMinorAlarm.setter
    def userUpperMinorAlarm(self, value):
        """
        The user-defined upper minor alarm for the scale.
        Parameters
        ----------
        value : float
            The new lower minor alarm value.
        """

        if self._minor_alarm_from_channel:
            return

        self._user_upper_minor_alarm = value
        self.scale_indicator.set_upper_minor_alarm(self._user_upper_minor_alarm)
        self.update_labels()

    @Property(float)
    def userLowerMinorAlarm(self):
        """
        The user-defined lower minor alarm for the scale.
        Returns
        -------
        float
        """
        return self._user_lower_minor_alarm

    @userLowerMinorAlarm.setter
    def userLowerMinorAlarm(self, value):
        """
        The user-defined lower minor alarm for the scale.
        Parameters
        ----------
        value : float
            The new lower minor alarm value.
        """

        if self._minor_alarm_from_channel:
            return

        self._user_lower_minor_alarm = value
        self.scale_indicator.set_lower_minor_alarm(self._user_lower_minor_alarm)
        self.update_labels()

    @Property(bool)
    def majorAlarmFromChannel(self):
        """
        Whether or not the scale indicator should use the major alarm information
        from the channel.
        Returns
        -------
        bool
        """
        return self._major_alarm_from_channel

    @majorAlarmFromChannel.setter
    def majorAlarmFromChannel(self, checked):
        """
        Whether or not the scale indicator should use the minor alarm information
        from the channel.
        Parameters
        ----------
        checked : bool
            True to use the major alarms from the Channel, False to use the user-defined
            values.
        """
        if self._major_alarm_from_channel != checked:
            self._major_alarm_from_channel = checked

            if checked:
                if self.lower_alarm_limit:
                    self.scale_indicator.set_lower_major_alarm(self.lower_alarm_limit)
                if self.upper_alarm_limit:
                    self.scale_indicator.set_upper_major_alarm(self.upper_alarm_limit)
            else:
                self.scale_indicator.set_lower_major_alarm(self._user_lower_major_alarm)
                self.scale_indicator.set_upper_major_alarm(self._user_upper_major_alarm)

            self.scale_indicator.set_lower_major_alarm(self.lower_alarm_limit)
            self.scale_indicator.set_upper_major_alarm(self.upper_alarm_limit)
            self.update_labels()

    @Property(float)
    def userUpperMajorAlarm(self):
        """
        The user-defined upper major alarm for the scale.
        Returns
        -------
        float
        """
        return self._user_upper_major_alarm

    @userUpperMajorAlarm.setter
    def userUpperMajorAlarm(self, value):
        """
        The user-defined upper major alarm for the scale.
        Parameters
        ----------
        value : float
            The new lower major alarm value.
        """

        if self._major_alarm_from_channel:
            return

        self._user_upper_major_alarm = value
        self.scale_indicator.set_upper_major_alarm(self._user_upper_major_alarm)

    @Property(float)
    def userLowerMajorAlarm(self):
        """
        The user-defined lower major alarm for the scale.
        Returns
        -------
        float
        """
        return self._user_lower_major_alarm

    @userLowerMajorAlarm.setter
    def userLowerMajorAlarm(self, value):
        """
        The user-defined lower major alarm for the scale.
        Parameters
        ----------
        value : float
            The new lower major alarm value.
        """

        if self._major_alarm_from_channel:
            return

        self._user_lower_major_alarm = value
        self.scale_indicator.set_lower_major_alarm(self._user_lower_major_alarm)

    # Two properties available on PyDMScaleIndicator we don't support for PyDMAnalogIndicator
    originAtZero = Property(bool, None, None, designable=False)
    barIndicator = Property(bool, None, None, designable=False)



import pytest
from unittest import mock
from pyqtgraph import AxisItem, PlotDataItem, ViewBox
from qtpy.QtWidgets import QGraphicsScene
from ...widgets.multi_axis_plot import MultiAxisPlot


@pytest.fixture
def sample_plot(qtbot, monkeypatch):
    """ Set up a plot with a couple of test axis items for use in test cases below """
    plot = MultiAxisPlot()
    scene = QGraphicsScene()
    monkeypatch.setattr(plot, 'scene', lambda: scene)  # Have the plot return a dummy scene when needed

    # Create two axes, one with a fixed range and one that should auto-range (adjust to always display all data)
    fixed_range_axis = AxisItem('left')
    plot.addAxis(fixed_range_axis, 'Fixed Range Axis', enableAutoRangeY=False, minRange=-5.0, maxRange=7.5)
    auto_range_axis = AxisItem('left')
    plot.addAxis(auto_range_axis, 'Auto Range Axis', enableAutoRangeY=True)

    return plot


def test_original_ranges_set_correctly(sample_plot):
    """ Check that the original ranges of x and y axes are preserved when adding them into the plot """

    # The axis with a set range should have that preserved
    assert sample_plot.axesOriginalRanges['Fixed Range Axis'][0] == -5.0
    assert sample_plot.axesOriginalRanges['Fixed Range Axis'][1] == 7.5

    # And the axis set to auto-range should be set to None indicating there is no fixed range to preserve
    assert sample_plot.axesOriginalRanges['Auto Range Axis'][0] is None
    assert sample_plot.axesOriginalRanges['Auto Range Axis'][1] is None


@mock.patch('pyqtgraph.ViewBox.enableAutoRange')
@mock.patch('pyqtgraph.PlotItem.setXRange')
@mock.patch('pyqtgraph.ViewBox.setYRange')
def test_restore_axis_ranges(mocked_y_range, mocked_x_range, mocked_auto_range, sample_plot):
    """
    Verify that when axis ranges have been changed from their defaults, calling restore will set them back to
    their original values
    """

    # First let's set a starting range on the bottom axis as well as those set in the fixture
    sample_plot.axesOriginalRanges['bottom'] = (0.0, 10.0)

    sample_plot.restoreAxisRanges()

    # After the call to restore axis ranges, ensure the view box calls are made as expected. (Don't want to test
    # view box functionality here as that is very involved and covered in pyqtgraph itself)
    mocked_auto_range.assert_any_call(axis=ViewBox.YAxis, enable=True)  # Auto-range enabled for our auto-range axis
    mocked_y_range.assert_called_once_with(-5.0, 7.5)  # Fixed range restored for the other axis
    mocked_x_range.assert_called_once_with(0.0, 10.0, padding=0)  # X-axis restored to the values specified above


def test_link_data_to_logarithmic_axis(qtbot, monkeypatch, sample_plot):
    """ Verify that when a curve is added to an axis with log mode on, that curve is set to log mode as well """

    # Create one linear axis and one logarithmic axis
    linear_axis = AxisItem('left')
    log_axis = AxisItem('left')
    log_axis.setLogMode(True)

    sample_plot.addAxis(linear_axis, 'Linear Axis')
    sample_plot.addAxis(log_axis, 'Log Axis')

    # Create a data item to go along with each axis
    linear_data = PlotDataItem()
    log_data = PlotDataItem()

    # Upon creation, all data should default to non-log mode
    assert linear_data.opts['logMode'] == [False, False]
    assert log_data.opts['logMode'] == [False, False]

    sample_plot.linkDataToAxis(linear_data, 'Linear Axis')
    sample_plot.linkDataToAxis(log_data, 'Log Axis')

    # Now that we've linked the data to their associated axes, the log_data should have logMode set to true, while
    # linear data should still have it set to false
    assert linear_data.opts['logMode'] == [False, False]
    assert log_data.opts['logMode'] == [False, True]


def test_update_log_mode(qtbot, sample_plot):
    """ Verify toggling log mode on and off for the entire plot works as expected """
    data_item = PlotDataItem()
    sample_plot.addItem(data_item)

    # For a brand new plot, log mode defaults to false for everything. Verify this is the case.
    assert data_item.opts['logMode'] == [False, False]
    for axis in sample_plot.getAxes():
        assert not axis.logMode

    # Now set log mode on for y-values only. Verify this gets set correctly, and x-values are left alone.
    sample_plot.setLogMode(False, True)
    assert data_item.opts['logMode'] == [False, True]
    for axis in sample_plot.getAxes():
        if axis.orientation in ('bottom', 'top'):
            assert not axis.logMode
        elif axis.orientation in ('left', 'right'):
            assert axis.logMode
        else:
            raise ValueError(f'Invalid value for axis orientation: {axis.orientation}')

    # Now set log mode on for x-values only. Verify this gets set correctly, and y-values are left alone.
    sample_plot.setLogMode(True, False)
    assert data_item.opts['logMode'] == [True, False]
    for axis in sample_plot.getAxes():
        if axis.orientation in ('bottom', 'top'):
            assert axis.logMode
        elif axis.orientation in ('left', 'right'):
            assert not axis.logMode
        else:
            raise ValueError(f'Invalid value for axis orientation: {axis.orientation}')

    # Now set log mode on for everything. Verify this is set across all items.
    sample_plot.setLogMode(True, True)
    assert data_item.opts['logMode'] == [True, True]
    for axis in sample_plot.getAxes():
        assert axis.logMode

    # And finally return everything back to non-log mode. Verify all items are no longer in log mode for x or y.
    sample_plot.setLogMode(False, False)
    assert data_item.opts['logMode'] == [False, False]
    for axis in sample_plot.getAxes():
        assert not axis.logMode


def test_remove_item(qtbot, sample_plot: MultiAxisPlot):
    """ Verify that removing an item from the plot works as expected """

    # First create a couple of mock data items to plot, and add them to the plot
    data_item_one = PlotDataItem()
    data_item_two = PlotDataItem()

    sample_plot.addItem(data_item_one)
    sample_plot.addItem(data_item_two)
    linked_viewbox = data_item_one.getViewBox()

    assert len(sample_plot.items) == 2
    assert len(sample_plot.dataItems) == 2
    assert len(linked_viewbox.addedItems) == 2

    # Now remove them one at a time, and ensure they are cleared out correctly
    sample_plot.removeItem(data_item_one)
    assert len(sample_plot.items) == 1
    assert len(sample_plot.dataItems) == 1
    assert len(linked_viewbox.addedItems) == 1
    assert sample_plot.items[0] == data_item_two
    assert sample_plot.dataItems[0] == data_item_two
    assert linked_viewbox.addedItems[0] == data_item_two

    sample_plot.removeItem(data_item_two)
    assert len(sample_plot.items) == 0
    assert len(sample_plot.dataItems) == 0
    assert len(linked_viewbox.addedItems) == 0

    # Finally, re-add them and verify that clear() will remove both at the same time
    sample_plot.addItem(data_item_one)
    sample_plot.addItem(data_item_two)
    sample_plot.clear()
    assert len(sample_plot.items) == 0
    assert len(sample_plot.dataItems) == 0
    assert len(linked_viewbox.addedItems) == 0



from scipy import constants

UNITS = {'length':   {'m': 1,
                      'cm': constants.centi,
                      'mm': constants.milli,
                      'um': constants.micro,
                      'nm': constants.nano,
                      'pm': constants.pico,
                      'in': constants.inch,
                      'ft': constants.foot,
                      'yds': constants.yard,
                      },
         'time':    {'s': 1,
                     'ms': constants.milli,
                     'us': constants.micro,
                     'ns': constants.nano,
                     'ps': constants.pico,
                     'min': constants.minute,
                     'hr': constants.hour,
                     'weeks': constants.week,
                     'days': constants.day,
                     },
         'frequency': {'Hz': 1,
                       'kHz': constants.kilo,
                       'MHz': constants.mega,
                       'GHz': constants.giga,
                       'THz': constants.tera,
                       'mHz': constants.milli,
                       },
         'angle':   {'rad': 1,
                     'mrad': constants.milli,
                     'urad': constants.micro,
                     'nrad': constants.nano,
                     'degree': constants.degree,
                     'turn': 2*constants.pi,
                     },
         'voltage': {'V': 1,
                     'MV': constants.mega,
                     'kV': constants.kilo,
                     'mV': constants.milli,
                     'uV': constants.micro,
                     },
         'current': {'A': 1,
                     'MA': constants.mega,
                     'kA': constants.kilo,
                     'mA': constants.milli,
                     'uA': constants.micro,
                     'nA': constants.nano,
                     }
         }


def find_unittype(unit):
    """
    Find the type of a unit string.

    Parameters
    ----------
    unit : str
        The unit string

    Returns
    -------
    tp : str
        The unit type name or None if not found.
    """
    for tp in UNITS.keys():
        if unit in UNITS[tp].keys():
            return tp
    return None


def find_unit(unit):
    """
    Find the conversion of a unit string.

    Parameters
    ----------
    unit : str
        The unit string

    Returns
    -------
    float or None
        The unit value relative to the standard or None if not found.
    """
    tp = find_unittype(unit)
    if tp:
        return UNITS[tp][unit]
    else:
        return None


def convert(unit, desired):
    """
    Find the conversion rate of two different unit strings.

    Parameters
    ----------
    unit : str
        The current unit string
    desired : str
        The desired unit string

    Returns
    -------
    float or None
        The relation between unit and desired or None if not found.
    """
    current = find_unit(unit)
    final = find_unit(desired)

    if find_unittype(unit) != find_unittype(desired):
        return None

    if current and final:
        return current/final


def find_unit_options(unit):
    """
    Find the options for a given unit.

    Parameters
    ----------
    unit : str
        The unit string

    Returns
    -------
    list or None
        The list of similar units in crescent order or None if not found.
    """
    tp = find_unittype(unit)
    if tp:
        units = [choice for choice, _ in
                 sorted(UNITS[tp].items(), key=lambda x: 1/x[1])]
        return units
    else:
        return None
